# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from functools import partial
from typing import Any, Callable, Sequence, Union

from google.protobuf.duration_pb2 import Duration
from google.protobuf.timestamp_pb2 import Timestamp

__all__ = [
    "DATETIME_ISO8601_Z_FORMAT",
    "DATE_FORMATS",
    "DATETIME_FORMATS",
    "TimeDeltaLike",
    "date_to_int",
    "date_to_str",
    "datetime_to_str",
    "datetime_to_epoch_microseconds",
    "datetime_to_epoch_timedelta",
    "datetime_to_timestamp",
    "timedelta_to_duration",
    "to_date",
    "to_datetime",
    "to_timedelta",
    # these are only exposed until dazl.util.prim_types is removed
    "_parse",
    "_parse_nano_format",
]

TimeDeltaLike = Union[int, float, Decimal, str, timedelta]

DATETIME_ISO8601_Z_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

_NANOS_PER_MICROSECOND = 1000
_SECONDS_PER_DAY = 24 * 3600


def _parse(fmt: str, value: str) -> "datetime":
    return datetime.strptime(value, fmt).replace(tzinfo=timezone.utc)


def _parse_nano_format(value: str) -> "datetime":
    if value.endswith("Z"):
        return _parse(DATETIME_ISO8601_Z_FORMAT, value[0:-4] + "Z")
    else:
        raise ValueError("could not parse")


DATE_FORMATS: Sequence[Callable[[str], datetime]] = [
    partial(_parse, "%Y-%m-%d"),
    partial(_parse, "%Y.%m.%d"),
]

DATETIME_FORMATS: Sequence[Callable[[str], datetime]] = [
    partial(_parse, DATETIME_ISO8601_Z_FORMAT),
    _parse_nano_format,
    partial(_parse, "%Y-%m-%dT%H:%M:%S.%f"),
    partial(_parse, "%Y-%m-%d %H:%M:%S.%f"),
    partial(_parse, "%Y-%m-%dT%H:%M:%SZ"),
    partial(_parse, "%Y-%m-%dT%H:%M:%S"),
    partial(_parse, "%Y-%m-%d %H:%M:%S"),
]


def to_date(obj: "Any") -> date:
    """
    Convert any of the common wire representations of a ``date`` to a ``date``.
    """
    if isinstance(obj, date):
        return obj
    elif isinstance(obj, datetime):
        return obj.date()
    elif isinstance(obj, (int, float)):
        return date(1970, 1, 1) + timedelta(days=obj)
    elif isinstance(obj, str):
        for fmt in DATE_FORMATS:
            try:
                return fmt(obj).date()
            except ValueError:
                pass

    raise ValueError(f"Could not parse as a date: {obj!r}")


def to_datetime(obj: "Any") -> datetime:
    """
    Convert any of the common wire representations of a ``datetime`` to a ``datetime``.
    """
    if isinstance(obj, datetime):
        # trivially return datetime objects untouched
        return obj
    elif isinstance(obj, Timestamp):
        # straight from a Google Timestamp
        dt = obj.ToDatetime()
        return dt.replace(tzinfo=timezone.utc)
    elif isinstance(obj, (int, float)):
        # from the gRPC Ledger API; interpret as microseconds from the GMT epoch
        return datetime.utcfromtimestamp(obj / 1e6).replace(tzinfo=timezone.utc)
    elif isinstance(obj, str):
        # from the REST Ledger API or an end user; give preference to the unambiguous wire
        # format ISO8601, but otherwise try to parse in a variety of formats
        for fmt in DATETIME_FORMATS:
            try:
                return fmt(obj)
            except ValueError:
                pass

    raise ValueError(f"Could not parse as a datetime: {obj!r}")


def to_timedelta(obj: "Any") -> "timedelta":
    """
    Convert an object to a timedelta. If the object is a timedelta, it is returned; if it is a
    numeric type, the value is interpreted as seconds and that is returned.

    :param obj: The object to convert.
    :return: A timedelta.

    >>> to_timedelta(timedelta(minutes=1, seconds=2))
    timedelta(minutes=1, seconds=2)
    >>> to_timedelta(30)
    timedelta(seconds=30)
    >>> to_timedelta(30.0)
    timedelta(seconds=30)
    >>> to_timedelta(Decimal(30))
    timedelta(seconds=30)
    """
    if obj is None:
        raise ValueError("obj cannot be None")
    if isinstance(obj, timedelta):
        return obj
    elif isinstance(obj, (int, float)):
        return timedelta(seconds=obj)
    elif isinstance(obj, Decimal):
        return timedelta(seconds=float(obj))
    elif isinstance(obj, str):
        return timedelta(seconds=float(obj))
    else:
        raise TypeError("could not convert {obj!r} to timedelta")


def date_to_int(obj: "date") -> int:
    """
    Return the Python ``datetime.date`` as the number of days since 1970 Jan 1.
    """
    return (obj - date(1970, 1, 1)).days


def date_to_str(obj: "date") -> str:
    """
    Convert a date to a JSON string that represents a date.
    """
    return obj.isoformat()


def datetime_to_str(obj: "datetime") -> str:
    """
    Converts the object to an ISO8601 datetime string.

    :param obj:
        A datetime. If the datetime is "naive" (no timezone information), it is
        assumed to refer to UTC. If the datetime has timezone information, the
        datetime will be converted to UTC before serializing.
    :return: An ISO8601 string that represents the time in UTC.
    """
    if obj.tzinfo is not None and obj.tzinfo.utcoffset(obj) is not None:
        # aware time; translate to UTC
        obj = obj.astimezone(timezone.utc)
    obj = obj.replace(tzinfo=None)
    return obj.isoformat() + "Z"


def datetime_to_timestamp(obj: "datetime") -> "Timestamp":
    """
    Convert the ``datetime`` to a Protobuf ``google.protobuf.Timestamp``.

    :param obj:
        A datetime. If the datetime is "naive" (no timezone information), it is
        assumed to refer to UTC. If the datetime has timezone information, the
        datetime will be converted to UTC before serializing.
    :return: A :class:`Timestamp`.
    """
    td = datetime_to_epoch_timedelta(obj)
    ts = Timestamp()
    ts.seconds = td.seconds + td.days * _SECONDS_PER_DAY
    ts.nanos = td.microseconds * _NANOS_PER_MICROSECOND
    return ts


def datetime_to_epoch_microseconds(obj: "datetime") -> float:
    """
    Return the Python ``datetime.datetime`` as number of microseconds since
    1970 Jan 1 midnight, UTC.
    """
    td = datetime_to_epoch_timedelta(obj)
    return (td.days * 86400 + td.seconds) * 10**6 + td.microseconds


def datetime_to_epoch_timedelta(obj: "datetime") -> "timedelta":
    """
    Return the Python ``datetime.datetime`` as an elapsed time from
    1970 Jan 1 midnight, UTC.
    """
    if obj.tzinfo is not None and obj.tzinfo.utcoffset(obj) is not None:
        # aware time; translate to UTC
        obj = obj.astimezone(timezone.utc)
    obj = obj.replace(tzinfo=None)
    return obj - datetime(1970, 1, 1, 0, 0, 0)


def timedelta_to_duration(obj: "timedelta") -> "Duration":
    """
    Return the Python ``timestamp`` as a Protobuf ``google.protobuf.Duration``.
    """
    d = Duration()
    d.seconds = int(obj.total_seconds())
    d.nanos = obj.microseconds * 1000
    return d



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/value.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"com/daml/ledger/api/v1/value.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\x1bgoogle/protobuf/empty.proto\"\xa0\x05\n\x05Value\x12\x38\n\x06record\x18\x01 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.RecordH\x00R\x06record\x12;\n\x07variant\x18\x02 \x01(\x0b\x32\x1f.com.daml.ledger.api.v1.VariantH\x00R\x07variant\x12!\n\x0b\x63ontract_id\x18\x03 \x01(\tH\x00R\ncontractId\x12\x32\n\x04list\x18\x04 \x01(\x0b\x32\x1c.com.daml.ledger.api.v1.ListH\x00R\x04list\x12\x1a\n\x05int64\x18\x05 \x01(\x12\x42\x02\x30\x01H\x00R\x05int64\x12\x1a\n\x07numeric\x18\x06 \x01(\tH\x00R\x07numeric\x12\x14\n\x04text\x18\x08 \x01(\tH\x00R\x04text\x12\"\n\ttimestamp\x18\t \x01(\x10\x42\x02\x30\x01H\x00R\ttimestamp\x12\x16\n\x05party\x18\x0b \x01(\tH\x00R\x05party\x12\x14\n\x04\x62ool\x18\x0c \x01(\x08H\x00R\x04\x62ool\x12,\n\x04unit\x18\r \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00R\x04unit\x12\x14\n\x04\x64\x61te\x18\x0e \x01(\x05H\x00R\x04\x64\x61te\x12>\n\x08optional\x18\x0f \x01(\x0b\x32 .com.daml.ledger.api.v1.OptionalH\x00R\x08optional\x12/\n\x03map\x18\x10 \x01(\x0b\x32\x1b.com.daml.ledger.api.v1.MapH\x00R\x03map\x12\x32\n\x04\x65num\x18\x11 \x01(\x0b\x32\x1c.com.daml.ledger.api.v1.EnumH\x00R\x04\x65num\x12\x39\n\x07gen_map\x18\x12 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.GenMapH\x00R\x06genMapB\x05\n\x03Sum\"\x86\x01\n\x06Record\x12?\n\trecord_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\x08recordId\x12;\n\x06\x66ields\x18\x02 \x03(\x0b\x32#.com.daml.ledger.api.v1.RecordFieldR\x06\x66ields\"X\n\x0bRecordField\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x05value\"s\n\nIdentifier\x12\x1d\n\npackage_id\x18\x01 \x01(\tR\tpackageId\x12\x1f\n\x0bmodule_name\x18\x03 \x01(\tR\nmoduleName\x12\x1f\n\x0b\x65ntity_name\x18\x04 \x01(\tR\nentityNameJ\x04\x08\x02\x10\x03\"\xa3\x01\n\x07Variant\x12\x41\n\nvariant_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\tvariantId\x12 \n\x0b\x63onstructor\x18\x02 \x01(\tR\x0b\x63onstructor\x12\x33\n\x05value\x18\x03 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x05value\"e\n\x04\x45num\x12;\n\x07\x65num_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\x06\x65numId\x12 \n\x0b\x63onstructor\x18\x02 \x01(\tR\x0b\x63onstructor\"A\n\x04List\x12\x39\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x08\x65lements\"?\n\x08Optional\x12\x33\n\x05value\x18\x01 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x05value\"\x92\x01\n\x03Map\x12;\n\x07\x65ntries\x18\x01 \x03(\x0b\x32!.com.daml.ledger.api.v1.Map.EntryR\x07\x65ntries\x1aN\n\x05\x45ntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x05value\"\xb7\x01\n\x06GenMap\x12>\n\x07\x65ntries\x18\x01 \x03(\x0b\x32$.com.daml.ledger.api.v1.GenMap.EntryR\x07\x65ntries\x1am\n\x05\x45ntry\x12/\n\x03key\x18\x01 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x03key\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x05valueB\x89\x01\n\x16\x63om.daml.ledger.api.v1B\x0fValueOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_VALUE = DESCRIPTOR.message_types_by_name['Value']
_RECORD = DESCRIPTOR.message_types_by_name['Record']
_RECORDFIELD = DESCRIPTOR.message_types_by_name['RecordField']
_IDENTIFIER = DESCRIPTOR.message_types_by_name['Identifier']
_VARIANT = DESCRIPTOR.message_types_by_name['Variant']
_ENUM = DESCRIPTOR.message_types_by_name['Enum']
_LIST = DESCRIPTOR.message_types_by_name['List']
_OPTIONAL = DESCRIPTOR.message_types_by_name['Optional']
_MAP = DESCRIPTOR.message_types_by_name['Map']
_MAP_ENTRY = _MAP.nested_types_by_name['Entry']
_GENMAP = DESCRIPTOR.message_types_by_name['GenMap']
_GENMAP_ENTRY = _GENMAP.nested_types_by_name['Entry']
Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Value)
  })
_sym_db.RegisterMessage(Value)

Record = _reflection.GeneratedProtocolMessageType('Record', (_message.Message,), {
  'DESCRIPTOR' : _RECORD,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Record)
  })
_sym_db.RegisterMessage(Record)

RecordField = _reflection.GeneratedProtocolMessageType('RecordField', (_message.Message,), {
  'DESCRIPTOR' : _RECORDFIELD,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.RecordField)
  })
_sym_db.RegisterMessage(RecordField)

Identifier = _reflection.GeneratedProtocolMessageType('Identifier', (_message.Message,), {
  'DESCRIPTOR' : _IDENTIFIER,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Identifier)
  })
_sym_db.RegisterMessage(Identifier)

Variant = _reflection.GeneratedProtocolMessageType('Variant', (_message.Message,), {
  'DESCRIPTOR' : _VARIANT,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Variant)
  })
_sym_db.RegisterMessage(Variant)

Enum = _reflection.GeneratedProtocolMessageType('Enum', (_message.Message,), {
  'DESCRIPTOR' : _ENUM,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Enum)
  })
_sym_db.RegisterMessage(Enum)

List = _reflection.GeneratedProtocolMessageType('List', (_message.Message,), {
  'DESCRIPTOR' : _LIST,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.List)
  })
_sym_db.RegisterMessage(List)

Optional = _reflection.GeneratedProtocolMessageType('Optional', (_message.Message,), {
  'DESCRIPTOR' : _OPTIONAL,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Optional)
  })
_sym_db.RegisterMessage(Optional)

Map = _reflection.GeneratedProtocolMessageType('Map', (_message.Message,), {

  'Entry' : _reflection.GeneratedProtocolMessageType('Entry', (_message.Message,), {
    'DESCRIPTOR' : _MAP_ENTRY,
    '__module__' : 'com.daml.ledger.api.v1.value_pb2'
    # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Map.Entry)
    })
  ,
  'DESCRIPTOR' : _MAP,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Map)
  })
_sym_db.RegisterMessage(Map)
_sym_db.RegisterMessage(Map.Entry)

GenMap = _reflection.GeneratedProtocolMessageType('GenMap', (_message.Message,), {

  'Entry' : _reflection.GeneratedProtocolMessageType('Entry', (_message.Message,), {
    'DESCRIPTOR' : _GENMAP_ENTRY,
    '__module__' : 'com.daml.ledger.api.v1.value_pb2'
    # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GenMap.Entry)
    })
  ,
  'DESCRIPTOR' : _GENMAP,
  '__module__' : 'com.daml.ledger.api.v1.value_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GenMap)
  })
_sym_db.RegisterMessage(GenMap)
_sym_db.RegisterMessage(GenMap.Entry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\017ValueOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _VALUE.fields_by_name['int64']._options = None
  _VALUE.fields_by_name['int64']._serialized_options = b'0\001'
  _VALUE.fields_by_name['timestamp']._options = None
  _VALUE.fields_by_name['timestamp']._serialized_options = b'0\001'
  _VALUE._serialized_start=92
  _VALUE._serialized_end=764
  _RECORD._serialized_start=767
  _RECORD._serialized_end=901
  _RECORDFIELD._serialized_start=903
  _RECORDFIELD._serialized_end=991
  _IDENTIFIER._serialized_start=993
  _IDENTIFIER._serialized_end=1108
  _VARIANT._serialized_start=1111
  _VARIANT._serialized_end=1274
  _ENUM._serialized_start=1276
  _ENUM._serialized_end=1377
  _LIST._serialized_start=1379
  _LIST._serialized_end=1444
  _OPTIONAL._serialized_start=1446
  _OPTIONAL._serialized_end=1509
  _MAP._serialized_start=1512
  _MAP._serialized_end=1658
  _MAP_ENTRY._serialized_start=1580
  _MAP_ENTRY._serialized_end=1658
  _GENMAP._serialized_start=1661
  _GENMAP._serialized_end=1844
  _GENMAP_ENTRY._serialized_start=1735
  _GENMAP_ENTRY._serialized_end=1844
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Collection, Mapping, Optional

from ...collections import merge

__all__ = [
    "PyType",
    "ANY",
    "OPTIONAL",
    "ITERABLE",
    "MAPPING",
    "BOOL",
    "FLOAT",
    "INT",
    "BYTES",
    "STRING",
    "ENUM_DESCRIPTOR",
    "SCALAR_CONTAINER",
    "COMPOSITE_CONTAINER",
    "MAP_CONTAINER",
    "BOTTOM",
]


class PyType:
    """
    Contains information about a resolved type in an import context.
    """

    def __init__(
        self,
        *,
        package: "Optional[str]" = None,
        py_str: str,
        imports: "Optional[Mapping[str, Collection[str]]]" = None,
    ):
        self.py_str = py_str
        self._imports = merge(imports, {package: [py_str]} if package else None)

    def apply(self, *args: "PyType") -> "PyType":
        """
        Produce a new type that is the application of this generic type with the specified type
        parameter.

        :param args: The type argument(s) to apply.
        """
        if not args:
            return self

        return PyType(
            py_str=f"{self.py_str}[" + ", ".join(arg.py_str for arg in args) + f"]",
            imports=merge(self.imports, *(arg.imports for arg in args)),
        )

    @property
    def settable(self) -> bool:
        return self.py_str == self.py_str.lower() or self.py_str.startswith("_typing.Literal[")

    @property
    def imports(self) -> "Mapping[str, Collection[str]]":
        """
        The imports that are needed in order to have well-formed Python code that utilizes the
        string representation of the type as represented by this object.
        """
        return self._imports


ANY = PyType(py_str="_typing.Any")
OPTIONAL = PyType(py_str="_typing.Optional")
ITERABLE = PyType(py_str="_typing.Iterable")
MAPPING = PyType(py_str="_typing.Mapping")
BOOL = PyType(py_str="_builtins.bool")
FLOAT = PyType(py_str="_builtins.float")
INT = PyType(py_str="_builtins.int")
BYTES = PyType(py_str="_builtins.bytes")
STRING = PyType(py_str="_builtins.str")
ENUM_DESCRIPTOR = PyType(package="google.protobuf.descriptor", py_str="EnumDescriptor")
SCALAR_CONTAINER = PyType(
    package="google.protobuf.internal.containers", py_str="RepeatedScalarFieldContainer"
)
COMPOSITE_CONTAINER = PyType(
    package="google.protobuf.internal.containers", py_str="RepeatedCompositeFieldContainer"
)
MAP_CONTAINER = PyType(package="google.protobuf.internal.containers", py_str="MessageMap")
BOTTOM = PyType(py_str="_typing.NoReturn")



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
import sys
from types import MappingProxyType
from typing import Any, Collection, DefaultDict, Mapping, Optional, Set, TypeVar

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

__all__ = ["merge"]


class SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool:
        ...


K = TypeVar("K")
V = TypeVar("V", bound=SupportsLessThan)  # noqa: Y001


def merge(*m: "Optional[Mapping[K, Collection[V]]]") -> "Mapping[K, Collection[V]]":
    """
    Combine multiple key-values mappings into a single key-values mapping.
    """
    ret = defaultdict(set)  # type: DefaultDict[K, Set[V]]
    for mapping in m:
        if mapping:
            for k, v in mapping.items():
                ret[k].update(v)
    return MappingProxyType({k: tuple(sorted(v)) for k, v in ret.items()})



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import BytesIO
from typing import TYPE_CHECKING, AbstractSet, Collection, Mapping, Optional, Sequence
import warnings

from ..damlast.daml_lf_1 import PackageRef
from ..damlast.parse import parse_archive
from ..damlast.pkgfile import Dar, DarFile as _DarFile

if TYPE_CHECKING:
    from ..model.types_store import PackageProvider, PackageStore


def get_archives(contents: bytes) -> "Mapping[PackageRef, bytes]":
    """
    Attempt to parse the specified contents as either a DALF or a DAR, and return a mapping of
    package IDs to DALF bytes.
    This class is deprecated as it assumes and expose use of a :class:`PackageStore` and

    :class:`Type`, all of which are deprecated.

    :param contents: A byte array to attempt to parse.
    :return: Return a mapping from package ID to byte contents.
    """
    warnings.warn(
        "get_archives is deprecated. For DAR parsing, use dazl.damlast.DarFile instead. "
        "There is no replacement for DALF parsing.",
        DeprecationWarning,
        stacklevel=2,
    )

    if contents[0:4] == b"PK\03\04":
        with BytesIO(contents) as buf:
            with _DarFile(buf) as dar:
                # noinspection PyProtectedMember
                return {name: dar._zip.read(name) for name in dar._dalf_names()}

    raise Exception(
        "get_archives(...) only works on DAR files, and is additionally deprecated. "
        "Use dazl.damlast.DarFile instead."
    )


class DarFile(_DarFile):
    """
    Provides access to the contents of a .dar file.
    """

    def __init__(self, dar: "Dar"):
        super().__init__(dar)
        warnings.warn(
            "dazl.util.dar.DarFile is deprecated; please use dazl.damlast.DarFile instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    def read_metadata(self) -> "PackageStore":
        warnings.warn(
            "read_metadata is deprecated. For DAR parsing, use dazl.damlast.DarFile instead. "
            "There is no replacement for DALF parsing.",
            DeprecationWarning,
            stacklevel=2,
        )
        # noinspection PyProtectedMember
        from ..model.types_store import PackageStore
        from ..protocols.v1.pb_parse_metadata import _parse_daml_metadata_pb

        store = PackageStore.empty()

        for archive in self.archives():
            store.register_all(_parse_daml_metadata_pb(archive))
        return store

    def get_archives(self) -> "Mapping[PackageRef, bytes]":
        """
        Return a mapping from package ID to byte contents.
        """
        warnings.warn(
            "get_archives is deprecated; there is no replacement.", DeprecationWarning, stacklevel=2
        )
        return {PackageRef(a.hash): a.payload for a in self._pb_archives()}

    def get_dalf_names(self) -> "Sequence[PackageRef]":
        warnings.warn(
            "get_dalf_names is deprecated; there is no replacement.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self._dalf_names())

    def get_manifest(self) -> "Optional[Mapping[str, str]]":
        """
        Return the contents of the manifest of this DAR.
        """
        warnings.warn(
            "get_manifest() is deprecated; use manifest() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.manifest()

    def get_sdk_version(self) -> "Optional[str]":
        """
        Return the SDK version used to compile this dar (if this information is available).

        This method is deprecated; use dazl.damlast..DarFile
        """
        warnings.warn(
            "get_sdk_version() is deprecated; use sdk_versions() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.sdk_version()

    def get_package_ids(self) -> "AbstractSet[PackageRef]":
        """
        Return the set of package IDs from this DAR.
        """
        warnings.warn(
            "get_package_ids() is deprecated; please use package_ids() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.package_ids()

    def get_package_provider(self) -> "PackageProvider":
        warnings.warn(
            "get_package_provider() is deprecated; use the methods of the "
            "dazl.damlast.PackageProvider protocol as implemented on DarFile directly.",
            DeprecationWarning,
        )
        from ..model.types_store import MemoryPackageProvider

        return MemoryPackageProvider(self.get_archives())


def parse_dalf(contents: bytes) -> "PackageStore":
    warnings.warn(
        "parse_dalf is deprecated, as PackageStore and Type are deprecated.",
        DeprecationWarning,
        stacklevel=2,
    )
    # noinspection PyProtectedMember
    from .._gen.com.daml.daml_lf_1_14.daml_lf_pb2 import Archive
    from ..protocols.v1.pb_parse_metadata import _parse_daml_metadata_pb

    a = Archive()
    a.ParseFromString(contents)
    return _parse_daml_metadata_pb(parse_archive(PackageRef(a.hash), a.payload))


def get_dar_package_ids(dar: "Dar") -> "Collection[PackageRef]":
    warnings.warn(
        "dazl.util.dar.get_dar_package_ids is deprecated; "
        "please use dazl.damlast.get_dar_package_ids instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    # This is a local import to get around circular dependencies; also the replacement function
    # is the same name as our own
    from ..damlast.pkgfile import get_dar_package_ids

    return get_dar_package_ids(dar)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, typing as _typing

import grpc as _grpc
from grpc import aio as _grpc_aio

from .user_management_service_pb2 import CreateUserRequest, CreateUserResponse, DeleteUserRequest, DeleteUserResponse, GetUserRequest, GetUserResponse, GrantUserRightsRequest, GrantUserRightsResponse, ListUserRightsRequest, ListUserRightsResponse, ListUsersRequest, ListUsersResponse, RevokeUserRightsRequest, RevokeUserRightsResponse

__all__ = [
    "UserManagementServiceStub",
]


# noinspection PyPep8Naming,DuplicatedCode
class UserManagementServiceStub:
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc.Channel) -> _UserManagementServiceBlockingStub: ...  # type: ignore
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc_aio.Channel) -> _UserManagementServiceAsyncStub: ...  # type: ignore
    def CreateUser(self, __1: CreateUserRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[CreateUserResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, CreateUserResponse]]: ...
    def GetUser(self, __1: GetUserRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetUserResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetUserResponse]]: ...
    def DeleteUser(self, __1: DeleteUserRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[DeleteUserResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, DeleteUserResponse]]: ...
    def ListUsers(self, __1: ListUsersRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[ListUsersResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, ListUsersResponse]]: ...
    def GrantUserRights(self, __1: GrantUserRightsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GrantUserRightsResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GrantUserRightsResponse]]: ...
    def RevokeUserRights(self, __1: RevokeUserRightsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[RevokeUserRightsResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, RevokeUserRightsResponse]]: ...
    def ListUserRights(self, __1: ListUserRightsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[ListUserRightsResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, ListUserRightsResponse]]: ...

# noinspection PyPep8Naming,DuplicatedCode
class _UserManagementServiceBlockingStub(UserManagementServiceStub):
    def CreateUser(self, __1: CreateUserRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> CreateUserResponse: ...
    def GetUser(self, __1: GetUserRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetUserResponse: ...
    def DeleteUser(self, __1: DeleteUserRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> DeleteUserResponse: ...
    def ListUsers(self, __1: ListUsersRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> ListUsersResponse: ...
    def GrantUserRights(self, __1: GrantUserRightsRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GrantUserRightsResponse: ...
    def RevokeUserRights(self, __1: RevokeUserRightsRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> RevokeUserRightsResponse: ...
    def ListUserRights(self, __1: ListUserRightsRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> ListUserRightsResponse: ...

# noinspection PyPep8Naming,DuplicatedCode
class _UserManagementServiceAsyncStub(UserManagementServiceStub):
    def CreateUser(self, __1: CreateUserRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, CreateUserResponse]: ...
    def GetUser(self, __1: GetUserRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetUserResponse]: ...
    def DeleteUser(self, __1: DeleteUserRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, DeleteUserResponse]: ...
    def ListUsers(self, __1: ListUsersRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, ListUsersResponse]: ...
    def GrantUserRights(self, __1: GrantUserRightsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GrantUserRightsResponse]: ...
    def RevokeUserRights(self, __1: RevokeUserRightsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, RevokeUserRightsResponse]: ...
    def ListUserRights(self, __1: ListUserRightsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, ListUserRightsResponse]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
:mod:`dazl.ledger.config` — connection configuration
====================================================

This module contains configuration objects for a :class:`Connection`.

Normally you don't need to construct these objects directly; instead, simply call
:meth:`dazl.connect`, which contains the same options exposed by the :meth:`Config.create` function,
which in turn includes the options exposed by the subobjects of :class:`Config`.

Generally you should not modify a :class:`Config` object (or subobjects) that have already been
passed to a :class:`Connection`. There are some exceptions to this rule, though; for example, when
using tokens, you can simply assign ``Config.access.token`` to a new value.

Configuration options are broken up into three subobjects:

* :ref:`Access configuration <access-configuration>`: settings commonly found in Daml JWT tokens
  (party settings, ledger ID, application name)
* :ref:`SSL/TLS configuration <ssl-tls-configuration>`: settings for configuring TLS connections
* :ref:`URL Configuration <url-configuration>`: settings that determine the location of gRPC Ledger
  API or HTTP JSON API implementation

.. autoclass:: Config
    :members:

.. _access-configuration:

Access configuration
--------------------

The :class:`AccessConfig` protocol specifies how ``dazl`` identifies itself to a ledger. There are
two built-in mechanisms for this: **property-based access**, which is traditionally used with
ledgers that do NOT require authorization/authentication (typical in a local development scenario,
for example coding against a local sandbox), and **token-based access**, which is required for
ledgers that DO require authorization/authentication (typical in a production scenario and/or hosted
ledgers).

In **property-based access** (:class:`PropertyBasedAccessConfig`), the behavior differs depending on
whether you are connecting over the gRPC Ledger API or the HTTP JSON API:

* For the *gRPC Ledger API*, ``read_as`` and ``act_as`` are used as-is. ``ledger_id`` is
  defaulted to the value requested from the ledger (but only if not initially specified).
  The ``admin`` property is ignored and unused.
* For the *HTTP JSON API*, ``read_as``, ``act_as``, ``admin``, ``ledger_id``, and
  ``application_name`` are all used to generate an unsigned JWT locally. ``ledger_id`` MUST be
  supplied.

In **token-based access** (:class:`TokenBasedAccessConfig`), the value of the token completely
determines the parties that can be used, the ledger ID to connect to, and the name of the
application. ``AccessConfig.token`` can be overwritten at any time, and that value will be used for
all subsequent calls to the ledger. If your ledger requires authorization/authentication using
tokens, you _must_ use token-based access.

Although ``dazl`` does not currently refresh tokens automatically, you can update the token yourself
at any time:

.. code-block:: python

    async def main():
        async with dazl.connect(token=MY_INITIAL_TOKEN) as conn:
            task1 = asyncio.create_task(do_ledger_stuff(conn))
            task2 = asyncio.create_task(refresh_token(conn))

            await task1
            task2.cancel()

    async def do_ledger_stuff(conn):
        # use the connection normally to make ledger calls
        ...

    async def refresh(conn):
        while True:
            # sleep for one hour
            await asyncio.sleep(3600)
            conn.config.access.token = NEW_REFRESHED_TOKEN

Deeper support for token refreshing may be added in a future release.

.. autoclass:: AccessConfig()
    :members:

.. autofunction:: create_access

.. autoclass:: PropertyBasedAccessConfig()
    :members:

.. autoclass:: TokenBasedAccessConfig()
    :members:

.. _ssl-tls-configuration:

SSL/TLS configuration
---------------------

.. autoclass:: SSLConfig
    :members:

.. _url-configuration:

URL configuration
-----------------

The :class:`URLConfig` protocol specifies the ledger implementation that ``dazl`` connects to
(either gRPC Ledger API or HTTP JSON API).

.. autofunction:: create_url

.. autoclass:: URLConfig()
    :members:

"""

import itertools
import logging
from logging import Logger
from os import PathLike
from typing import Collection, Optional, Union

from ...damlast.lookup import MultiPackageLookup
from ...prim import Party, TimeDeltaLike
from .access import AccessConfig, PropertyBasedAccessConfig, TokenBasedAccessConfig, create_access
from .argv import configure_parser
from .ssl import SSLConfig
from .url import URLConfig, create_url

__all__ = [
    "Config",
    "AccessConfig",
    "SSLConfig",
    "URLConfig",
    "TokenBasedAccessConfig",
    "PropertyBasedAccessConfig",
    "create_access",
    "create_url",
    "configure_parser",
]

# an incrementing counter that helps keep loggers for individual config connections unique
# (see https://mail.python.org/pipermail//python-ideas/2016-August/041871.html); this is safe to
# do even in multithreaded environments because ultimately we're reusing the GIL as our lock
id_generator = itertools.count()


# PyCharm thinks ``from .url import ...`` clashes with variables named ``url``
# (same with ``ssl`` and ``access``).
# noinspection PyShadowingNames
class Config:
    """
    Stores configuration for a :class:`Connection`.
    """

    @classmethod
    def create(
        cls,
        url: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        scheme: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
        admin: Optional[bool] = False,
        ledger_id: Optional[str] = None,
        application_name: Optional[str] = None,
        oauth_token: Optional[str] = None,
        oauth_token_file: Optional[str] = None,
        ca: Optional[bytes] = None,
        ca_file: Optional[PathLike] = None,
        cert: Optional[bytes] = None,
        cert_file: Optional[PathLike] = None,
        cert_key: Optional[bytes] = None,
        cert_key_file: Optional[PathLike] = None,
        connect_timeout: Optional[TimeDeltaLike] = None,
        use_http_proxy: bool = True,
        logger: Optional[Logger] = None,
        logger_name: Optional[str] = None,
        log_level: Optional[str] = None,
        lookup: "Optional[MultiPackageLookup]" = None,
    ) -> "Config":
        """
        Create a :class:`Config` object from the supplied parameters.

        The remote can be configured either by supplying a URL or by supplying a host (and optional
        port and scheme). If none of ``url``, ``host``, ``port``, or ``scheme`` are supplied, then
        environment variables are consulted; if no environment variables are specified either, then
        default values are used.

        The remote can be configured either by supplying a URL or by supplying a host (and optional
        port and scheme). If none of ``url``, ``host``, ``port``, or ``scheme`` are supplied, then
        environment variables are consulted; if no environment variables are specified either, then
        default values are used.

        When the URL scheme is ``http`` or ``https``, dazl will first attempt to connect assuming
        the HTTP JSON API; if this fails, gRPC Ledger API is attempted.

        If ``oauth_token`` is supplied and non-empty, then *token-based access* is used to connect
        to the ledger.

        If other fields are specified, then *property-based access* is used. At least one of
        ``read_as``, ``act_as``, or ``admin`` must be supplied. For the HTTP JSON API,
        ``ledger_id`` MUST be supplied.

        If _no_ fields are specified, this is an error unless environment variables supply an
        alternate source of configuration.

        +-------------------+----------------------+-----------------------------------------------+
        | function argument | environment variable | default value                                 |
        +===================+======================+===============================================+
        | ``url``           | DAML_LEDGER_URL      | localhost:6865                                |
        +-------------------+----------------------+-----------------------------------------------+
        | ``host``          | DAML_LEDGER_HOST     | localhost                                     |
        +-------------------+----------------------+-----------------------------------------------+
        | ``port``          | DAML_LEDGER_PORT     | 6865, unless ``scheme`` is specified:         |
        |                   |                      | * 80 for ``http``                             |
        |                   |                      | * 443 for ``https``                           |
        |                   |                      | * 6865 for ``grpc``                           |
        +-------------------+----------------------+-----------------------------------------------+
        | ``scheme``        | DAML_LEDGER_SCHEME   | | ``https`` for port 443 or 8443              |
        |                   |                      | | ``http`` for port 80, 7575 or 8080          |
        |                   |                      | | ``grpc`` for port 6865                      |
        |                   |                      | | ``https`` for all other ports               |
        +-------------------+----------------------+-----------------------------------------------+

        **HTTP(s) proxies (gRPC Ledger API only)**

        When connecting to the gRPC Ledger API, note that
        `gRPC environment variables
        <https://grpc.github.io/grpc/cpp/md_doc_environment_variables.html>`_
        are always also respected; you can set ``https_proxy``/`http_proxy`` (note the lowercase
        environment variable name). dazl, by default, will _also_ disable usage of a proxy server
        for a ``localhost`` host or a ``127.0.0.1`` host; this can be overridden by passing a value
        of ``use_http_proxy`` to ``True``.

        :param url:
            The URL to connect to. Can be used as an alternative to supplying ``host``, ``port``,
            and ``scheme`` as individual values. Can alternatively be supplied via the environment
            variable ``DAML_LEDGER_URL``.
        :param host:
            The host to connect to. Can be used as an alternative to supplying ``url`` as a combined
            value. Can alternatively be supplied via the environment variable ``DAML_LEDGER_HOST``.
        :param port:
            The port to connect to. Can be used as an alternative to supplying ``url`` as a combined
            value. Can alternatively be supplied via the environment variable ``DAML_LEDGER_PORT``.
        :param scheme:
            The scheme to connect to. Can be used as an alternative to supplying ``url`` as a
            combined value. Can alternatively be supplied via the environment variable
            ``DAML_LEDGER_SCHEME``.
        :param connect_timeout:
            Length of time to wait before giving up connecting to the remote and declaring an error.
            The default value is 30 seconds.
        :param use_http_proxy:
            ``True`` to use an HTTP(S) proxy server if configured; ``False`` to avoid using any
            configured server. If unspecified and the host is ``localhost``, the proxy server is
            avoided; otherwise the proxy server is used.
        :param ca:
            A certificate authority to use to validate the server's certificate. If not supplied,
            the operating system's default trust store is used. Cannot be specified with
            ``ca_file``.
        :param ca_file:
            A file containing the certificate authority to use to validate the server's certificate.
            If not supplied, the operating system's default trust store is used. Cannot be specified
            with ``ca``.
        :param cert:
            A client-side certificate to be used when connecting to a server that requires mutual
            TLS. Cannot be specified with ``cert_file``.
        :param cert_file:
            A file containing the client-side certificate to be used when connecting to a server
            that requires mutual TLS. Cannot be specified with ``cert``.
        :param cert_key:
            A client-side private key to be used when connecting to a server that requires mutual
            TLS. Cannot be specified with ``cert_key_file``.
        :param cert_key_file:
            A client-side private key to be used when connecting to a server that requires mutual
            TLS. Cannot be specified with ``cert_key``.
        :param read_as:
            A party or set of parties on whose behalf (in addition to all parties listed in
            ``act_as``) contracts can be retrieved. Cannot be specified if ``oauth_token`` or
            ``oauth_token_file`` is specified.
        :param act_as:
            A party or set of parties on whose behalf commands should be executed. Parties here are
            also implicitly granted ``read_as`` access as well. Cannot be specified if
            ``oauth_token`` or ``oauth_token_file`` is specified.
        :param admin:
            HTTP JSON API only: allow admin endpoints to be used. This flag is ignored when
            connecting to gRPC Ledger API implementations. Cannot be specified if ``oauth_token`` or
            ``oauth_token_file`` is specified.
        :param ledger_id:
            The ledger ID to connect to. For the HTTP JSON API, this value is required. For the gRPC
            Ledger API, if this value is _not_ supplied, its value will be retrieved from the
            server. Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.
        :param application_name:
            A string that identifies this application. This is used for tracing purposes on the
            server-side. Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is
            specified.
        :param oauth_token:
            The OAuth bearer token to be used on all requests. If specified, no other access
            parameters can be specified.
        :param oauth_token_file:
            A file that contains the OAuth bearer token to be used on all requests. If specified, no
            other access parameters can be specified.
        :param logger:
            The logger to use for connections created from this configuration. If not supplied, a
            logger will be created.
        :param logger_name:
            The name of the logger. Only used if ``logger`` is not provided.
        :param log_level:
            The logging level for the logger. The default is ``warn``. Only used if ``logger`` is
            not provided.
        :param lookup:
            An alternate symbol table to use to store package information. You should not normally
            need to set this value.
        """
        if logger is None:
            if not logger_name:
                logger_name = f"dazl.conn.{next(id_generator)}"
            logger = logging.getLogger(logger_name)
            if log_level is not None:
                logger.setLevel(log_level)

        url_config = create_url(
            url=url,
            host=host,
            port=port,
            scheme=scheme,
            connect_timeout=connect_timeout,
            use_http_proxy=use_http_proxy,
            logger=logger,
        )

        access_config = create_access(
            read_as=read_as,
            act_as=act_as,
            admin=admin,
            ledger_id=ledger_id,
            application_name=application_name,
            oauth_token=oauth_token,
            oauth_token_file=oauth_token_file,
            logger=logger,
        )

        ssl_config = SSLConfig(
            ca=ca,
            ca_file=ca_file,
            cert=cert,
            cert_file=cert_file,
            cert_key=cert_key,
            cert_key_file=cert_key_file,
            logger=logger,
        )

        return cls(access_config, ssl_config, url_config, logger, lookup)

    def __init__(
        self,
        access: AccessConfig,
        ssl: SSLConfig,
        url: URLConfig,
        logger: Logger,
        lookup: "Optional[MultiPackageLookup]" = None,
    ):
        """
        Initialize an instance of :class:`Config`.
        """
        self.access = access
        self.ssl = ssl
        self.url = url
        self.logger = logger
        self.lookup = lookup



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# type: ignore

from io import StringIO
from typing import Mapping, Optional, Sequence, Union
import warnings

from .. import LOG
from ..damlast import daml_types as daml
from ..damlast.daml_lf_1 import (
    Block,
    Case,
    DefDataType,
    DefTemplate,
    Expr,
    PrimType,
    Scenario,
    Type,
    TypeConName,
    Update,
)
from ..damlast.util import def_value, module_name
from ._render_base import PrettyPrintBase, pretty_print_syntax
from .util import indent, maybe_parentheses

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from ..model.types import ModuleRef, Type as OldType
    from ..model.types_store import PackageStore


def values_by_module(
    store: "PackageStore",
) -> "Mapping[ModuleRef, Mapping[Sequence[str], Union[Expr, OldType]]]":
    from collections import defaultdict

    d = defaultdict(defaultdict)
    for vn, vv in store._value_types.items():
        d[vn.module][vn.name] = vv
    for vn, vv in store._data_types.items():
        d[vn.module][vn.name] = vv
    return d


@pretty_print_syntax("python")
class PythonPrettyPrint(PrettyPrintBase):
    """
    Renderer for :class:`Expr` and :class:`Type` hierarchies that output text that look like
    Python code.
    """

    syntax = "python"

    def lexer(self):
        try:
            from pygments.lexers.python import PythonLexer
        except ImportError:
            return None
        return PythonLexer()

    def visit_module_ref_start(self, module_ref: "ModuleRef") -> str:
        mn = str(module_name(module_ref))
        _, _, last_component = mn.rpartition(".")

        return f"@module\nclass {last_component}:\n" if last_component else ""

    # def visit_def_value(self, def_value: 'DefValue', as_instance_method: bool = False):
    #     def_value.name_with_type

    def visit_empty_block_body(self):
        return "pass"

    def visit_def_template(
        self, template: "Optional[DefTemplate]", def_data_type: "Optional[DefDataType]" = None
    ) -> str:
        template_full_name = str(
            TypeConName(self.context.current_module, def_data_type.name.segments)
        )
        template_name = def_data_type.name.segments[-1]
        slot_names = (
            tuple(fwt.field for fwt in def_data_type.record.fields) if def_data_type.record else ()
        )
        signatories_def = (
            def_value("signatories", daml.List(daml.Party), template.signatories)
            if template is not None
            else None
        )

        if template is not None:
            lines = []
            lines.append(
                f"class {template_name}(metaclass=TemplateMeta, template_name={template_full_name!r}):"
            )
            lines.append('    """')
            lines.append("    Example usage:")
            lines.append(f'        create({template_full_name.replace(":", ".")},')
            lines.append(f"               {python_example_object(self.store, def_data_type)})")
            for choice in template.choices:
                lines.append(
                    f"        exercise(cid, {choice.name!r}, {python_example_object(self.store, choice.arg_binder.type)})"
                )
            lines.append('    """')
            lines.append(indent(self._visit_def_type_body(slot_names), 4))
            for choice in template.choices:
                ct = choice.arg_binder.type
                if ct.prim is not None and ct.prim.prim == PrimType.UNIT:
                    choice_slot_names = ()
                else:
                    tt = self.store.resolve_type_reference(ct.con.tycon)
                    choice_slot_names = (
                        tuple(name for name, _ in tt.named_args) if tt is not None else ()
                    )
                lines.append(
                    f"    class {choice.name}(metaclass=ChoiceMeta, template_name={template_full_name!r}, choice_name={choice.name!r}):"
                )
                lines.append(indent(self._visit_def_type_body(choice_slot_names), 8))
            return "\n".join(lines)
        else:
            return ""

    def _visit_def_type_body(self, slot_names: "Sequence[str]") -> str:
        lines = [f"__slots__ = {slot_names!r}\n"]
        if slot_names:
            lines.append(f'def __init__(self, {", ".join(slot_names)}):')
            lines.extend(f"    self.{slot_name} = {slot_name}" for slot_name in slot_names)
            lines.append("")
            lines.append(f"def _asdict(self) -> dict:")
            lines.append(f"    return {{")
            lines.extend(f"        {slot_name!r}: self.{slot_name}," for slot_name in slot_names)
            lines.append(f"        }}")
        else:
            lines.append("def _asdict(self) -> dict:")
            lines.append("    return {}")
        return "\n".join(lines) + "\n"

    def visit_expr_prim_con_inline_unit(self):
        return "{}"

    def visit_expr_rec_con_inline(self, rec_con: "Expr.RecCon"):
        with StringIO() as buf:
            buf.write("record(")
            delim = ""
            for fwe in rec_con.fields:
                buf.write(delim)
                buf.write(fwe.field)
                buf.write("=")
                expr = self.visit_expr(fwe.expr)
                if " " in expr:
                    buf.write("(")
                buf.write(expr)
                if " " in expr:
                    buf.write(")")
                delim = ", "
            buf.write(")")

            # if self.decl_name is not None:
            #     buf.write('  # type: ')
            #     buf.write(self.visit_type_con(rec_con.tycon))
            return buf.getvalue()

    def visit_expr_rec_proj(self, rec_proj: "Expr.RecProj") -> str:
        record_text = self.visit_expr(rec_proj.record)
        if " " in record_text:
            return f"({record_text})[{rec_proj.field!r}]"
        else:
            return f"{record_text}[{rec_proj.field!r}]"

    def visit_expr_variant_con(self, variant_con: "Expr.VariantCon") -> str:
        with StringIO() as buf:
            try:
                scope = self
                if self.context.decl_name is not None:
                    buf.write(".".join(self.context.decl_name))
                    buf.write(" = ")
                else:
                    buf.write("return ")
                buf.write("variant(")
                buf.write(variant_con.variant_con)
                buf.write("=")
                expr = scope.visit_expr(variant_con.variant_arg)
                if " " in expr:
                    buf.write("(")
                buf.write(expr)
                if " " in expr:
                    buf.write(")")
                buf.write(")")

                if self.context.decl_name is not None:
                    buf.write("  # type: ")
                    buf.write(self.visit_type_con(variant_con.tycon))
                return buf.getvalue()
            except:
                LOG.exception("why why why")
                exit(-1)

    def visit_expr_struct_con(self, struct_con: "Expr.StructCon") -> str:
        if struct_con.fields:
            with StringIO() as buf:
                delim = "tuple("
                for fwe in struct_con.fields:
                    buf.write(delim)
                    buf.write(fwe.field)
                    buf.write("=")
                    expr = self.visit_expr(fwe.expr)
                    if " " in expr:
                        buf.write("(")
                    buf.write(expr)
                    if " " in expr:
                        buf.write(")")
                    delim = ", "
                buf.write(")")
                return buf.getvalue()
        else:
            return "tuple()"

    def visit_expr_struct_proj(self, struct_proj: "Expr.StructProj") -> str:
        record_text = self.visit_expr(struct_proj.tuple)
        if " " in record_text:
            return f"({record_text})[{struct_proj.field!r}]"
        else:
            return f"{record_text}[{struct_proj.field!r}]"

    def visit_expr_app_inline(self, app: "Expr.App"):
        fun_text = maybe_parentheses(self.visit_expr(app.fun))
        args_text = ", ".join(self.visit_expr(arg) for arg in app.args)
        return f"{fun_text}({args_text})"

    def visit_expr_ty_app(self, ty_app: "Expr.TyApp") -> str:
        fun_text = self.visit_expr(ty_app.expr)
        return fun_text
        args_text = ", ".join(self.visit_type(arg) for arg in ty_app.types)
        return f"ty_apply({fun_text})<{args_text}>"

    def visit_expr_abs_decl(self, abs_: "Expr.Abs", context: "CodeContext"):
        body = self.visit_expr(abs_.body)
        try:
            body_type = self.type_computer.visit_expr(abs_.body)
            body_type_expr = " -> '" + self.visit_type(body_type) + "'"
        except:
            # LOG.exception("Couldn't compute a type!")
            body_type_expr = ""
        safe_decl_name = ".".join(context.decl_name).replace("$", "_")
        args = ", ".join(f"{p.var}: '{self.visit_type(p.type)}'" for p in abs_.param)
        decl_line = f"@staticmethod\ndef {safe_decl_name}({args}){body_type_expr}:\n"
        return decl_line + "".join(("    " + line) for line in body.splitlines(keepends=True))

    def visit_expr_abs_inline(self, abs_: "Expr.Abs"):
        body = self.visit_expr(abs_.body)
        args = ", ".join(f"{p.var}" for p in abs_.param)
        if args:
            args = " " + args
        return f"lambda{args}: {body}"

    def visit_expr_ty_abs(self, ty_abs: "Expr.TyAbs") -> str:
        args = ", ".join(p.var for p in ty_abs.param)
        if args:
            args = " " + args
        body = self.visit_expr(ty_abs.body)
        return body

        return f"ty_lambda{args}: {body}"

    def visit_expr_case(self, case: "Case") -> str:
        scrut = self.visit_expr(case.scrut)
        line = f"    scrut = {scrut}\n"
        # for alt in case.alts:
        #     alt.Sum_match()
        #     alt.body
        line += "    ..."
        return line

    def visit_expr_let(self, let: "Block") -> str:
        if self.context.in_expression:
            assignments = ", ".join(
                f"{b.binder.var} := {self.visit_expr(b.bound)}" for b in let.bindings
            )
            return f"({assignments}, {self.visit_expr(let.body)})[-1]"
        else:
            with StringIO() as buf:
                for b in let.bindings:
                    buf.write(f"{b.binder.var} = {self.visit_expr(b.bound)}\n")
                buf.write(f"return {self.visit_expr(let.body)}")
                return buf.getvalue()

    def visit_expr_nil_inline(self, nil: "Expr.Nil") -> str:
        return "()"

    def visit_expr_cons(self, cons: "Expr.Cons") -> str:
        with StringIO() as buf:
            delim = ""
            buf.write("[")
            for f in cons.front:
                buf.write(delim)
                expr = self.visit_expr(f)
                if " " in expr:
                    buf.write("(")
                buf.write(expr)
                if " " in expr:
                    buf.write(")")
                delim = ", "

            buf.write(delim)
            expr = self.visit_expr(cons.tail)
            buf.write("*")
            if " " in expr:
                buf.write("(")
            buf.write(expr)
            if " " in expr:
                buf.write(")")
            buf.write("]")
            return buf.getvalue()

    def visit_expr_update(self, update: "Update") -> str:
        raise Exception

    def visit_expr_scenario(self, scenario: "Scenario") -> str:
        raise Exception

    def visit_expr_rec_upd(self, rec_upd: "Expr.RecUpd") -> str:
        raise Exception

    def visit_expr_struct_upd(self, struct_upd: "Expr.StructUpd") -> str:
        raise Exception

    def visit_expr_optional_none(self, optional_none: "Expr.OptionalNone") -> str:
        expr = "None"
        if self.context.in_expression:
            return expr
        else:
            try:
                type_str = self.visit_type(optional_none.type)
                type_str = f"  # type: Optional[{type_str}]"
            except:
                type_str = None

            return f"return {expr}{type_str}"

    def visit_expr_optional_some(self, optional_some: "Expr.OptionalSome") -> str:
        expr = self.visit_expr(optional_some.body)
        if self.in_lambda_body:
            return expr
        else:
            try:
                type_str = self.visit_type(optional_some.type)
                type_str = f"  # type: Optional[{type_str}]"
            except:
                type_str = None

            return f"return {expr}{type_str}"

    def visit_type_var(self, var: "Type.Var") -> str:
        return var.var

    def visit_type_con(self, con: "Type.Con") -> str:
        if con.args:
            return f'{con.tycon.full_name}[{", ".join([self.visit_type(a) for a in con.args])}]'
        else:
            return con.tycon.full_name

    def visit_type_prim(self, prim: "Type.Prim") -> str:
        if PrimType.UNIT == prim.prim:
            return "Unit"
        elif PrimType.BOOL == prim.prim:
            return "bool"
        elif PrimType.INT64 == prim.prim:
            return "int"
        elif PrimType.DECIMAL == prim.prim:
            return "Decimal"
        elif PrimType.CHAR == prim.prim:
            return "str"
        elif PrimType.TEXT == prim.prim:
            return "str"
        elif PrimType.TIMESTAMP == prim.prim:
            return "datetime"
        elif PrimType.RELTIME == prim.prim:
            return "timedelta"
        elif PrimType.PARTY == prim.prim:
            return "Party"
        elif PrimType.LIST == prim.prim:
            if len(prim.args) == 0:
                return "List"
            else:
                return f"List[{prim.args[0]}]"
        elif PrimType.UPDATE == prim.prim:
            if len(prim.args) == 0:
                return "Update"
            else:
                return f"Update[{self.visit_type(prim.args[0])}]"
        elif PrimType.SCENARIO == prim.prim:
            if len(prim.args) == 0:
                return "Scenario"
            else:
                return f"Scenario[{self.visit_type(prim.args[0])}]"
        elif PrimType.DATE == prim.prim:
            return "Date"
        elif PrimType.CONTRACT_ID == prim.prim:
            return f"ContractId[{self.visit_type(prim.args[0])}]"
        elif PrimType.OPTIONAL == prim.prim:
            if len(prim.args) == 0:
                return "Optional"
            else:
                return f"Optional[{self.visit_type(prim.args[0])}]"
        elif PrimType.ARROW == prim.prim:
            type_strings = [self.visit_type(a) for a in prim.args]
            return f'Callable[[{", ".join(type_strings[:-1])}], {type_strings[-1]}]'
        elif PrimType.TEXTMAP == prim.prim:
            return f"Map[{self.visit_type(prim.args[0])},  {self.visit_type(prim.args[1])}]"
        elif PrimType.TYPE_REP == prim.prim:
            return "???"
        else:
            raise ValueError(f"unknown Type.Prim: {prim!r}")

    def visit_type_forall(self, forall: "Type.Forall") -> str:
        # Python does not have a way of modelling existential quantification, so simply throw this
        # information away
        return self.visit_type(forall.body)

    def visit_type_struct(self, struct: "Type.Struct") -> str:
        return "?"


def python_example_object(*_, **__):
    return "..."



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

__all__ = ["ConfigError", "ConfigWarning"]


class ConfigError(ValueError):
    pass


class ConfigWarning(Warning):
    pass



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
from itertools import product
import json
from keyword import iskeyword
import os.path
from typing import Dict, TextIO

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse
from google.protobuf.descriptor_pb2 import DescriptorProto, EnumDescriptorProto, FileDescriptorProto

from ..syntax.python import ImportContext, SymbolTable, Usage, all_decl, py_message_package
from ._header import HEADER

__all__ = ["run_plugin", "typing_file"]


def run_plugin(request: "CodeGeneratorRequest") -> "CodeGeneratorResponse":
    """
    Generate .pyi (typing stubs files) for Protobuf objects.
    """
    # build a mapping of all types that we'll ever need to generate any of the requested files
    symbol_table = SymbolTable()
    for fd in request.proto_file:
        symbol_table.load_file(fd)

    return CodeGeneratorResponse(
        file=[
            typing_file(pf, ImportContext(symbol_table))
            for pf in request.proto_file
            if pf.name in request.file_to_generate
        ]
    )


def typing_file(fd: "FileDescriptorProto", ictx: "ImportContext") -> "CodeGeneratorResponse.File":
    name = os.path.splitext(fd.name)[0] + "_pb2.pyi"

    # first, generate all message and enum bodies
    with StringIO() as buf:
        for ed in fd.enum_type:
            write_enum(buf, ed, ictx)

        for md in fd.message_type:
            write_message(buf, md, ictx)

        body = buf.getvalue()

    # once we know all of the enums/messages that we want to write, put together the set of imports
    ictx.add_system_import("", "builtins as _builtins")
    ictx.add_system_import("", "sys")
    ictx.add_system_import("", "typing as _typing")

    imports = ictx.py_import_block(py_message_package(fd.name))
    imports += (
        "\nif sys.version_info >= (3, 8):\n"
        "    from typing import Literal as _L\n"
        "else:\n"
        "    from typing_extensions import Literal as _L\n"
    )

    all_str = all_decl(md.name for md in fd.message_type)

    return CodeGeneratorResponse.File(name=name, content=f"{HEADER}\n{imports}\n{all_str}\n{body}")


def write_enum(buf: "TextIO", d: "EnumDescriptorProto", ictx: "ImportContext") -> None:
    ictx.add_import("google.protobuf.descriptor", "EnumDescriptor")

    buf.write(f"class {d.name}:\n")
    buf.write(f"    DESCRIPTOR: _typing.ClassVar[EnumDescriptor] = ...\n")
    for v in d.value:
        buf.write(f"    {v.name}: _typing.ClassVar[_L[{v.number}]] = ...\n")
    for v in d.value:
        buf.write(f"{v.name} = _L[{v.number}]\n")
    buf.write("\n")


def write_message(buf: "TextIO", md: "DescriptorProto", ictx: "ImportContext") -> None:
    if md.options.map_entry:
        # map types are represented as auto-generated messages;
        # there is nothing to emit for these types
        return

    ictx.add_import("google.protobuf.message", "Message as _Message")

    buf.write("\n")
    buf.write(f"class {md.name}(_Message):\n")

    # Write any nested messages first.
    if md.nested_type:
        with StringIO() as child_buf:
            for ed in md.enum_type:
                write_enum(child_buf, ed, ictx)
            for nd in md.nested_type:
                write_message(child_buf, nd, ictx)

            for line in child_buf.getvalue().splitlines():
                buf.write(f"    {line}\n")

    # Write fields.
    for fld in md.field:
        if not iskeyword(fld.name):
            py_type = ictx.py_type(fld, Usage.GETTER)
            if py_type.settable:
                buf.write(f"    {fld.name}: {py_type.py_str}\n")
            else:
                buf.write(f"    @property\n")
                buf.write(f"    def {fld.name}(self) -> {py_type.py_str}: ...\n")

    # for each oneof group that we have, generate a list of field names that represent the possible
    # options; also include an empty string which indicates the constructor that leaves out this
    # argument altogether
    oneof_groups = {
        od.name: [
            "",
            *(fld.name for fld in md.field if fld.HasField("oneof_index") and fld.oneof_index == i),
        ]
        for i, od in enumerate(md.oneof_decl)
    }

    # The constructor cannot include any arguments that are Python keywords. Callers can use kwargs
    # to pass in that information, but unfortunately there does not appear to be a good way to
    # expose that information, even through pyi files.
    #
    # Consequently we only generate permutations of constructors for combinations where arguments
    # are not Python keywords.
    oneof_field_names_groups = list(
        product(*[[f for f in g if not iskeyword(f)] for g in oneof_groups.values()])
    )
    for oneof_field_names in oneof_field_names_groups:
        init_fields = [
            fld
            for fld in md.field
            if not iskeyword(fld.name)
            and (not fld.HasField("oneof_index") or oneof_field_names[fld.oneof_index] == fld.name)
        ]
        self_name = "self"
        while self_name in [f.name for f in init_fields]:
            self_name += "_"

        if len(oneof_field_names_groups) > 1:
            buf.write("    @_typing.overload\n")

        if init_fields:
            buf.write(f"    def __init__({self_name}, *")

            for fld in init_fields:
                py_type = ictx.py_type(fld, Usage.INIT)
                buf.write(f", {fld.name}: {py_type.py_str} = ...")
            buf.write("): ...\n")
        else:
            buf.write(f"    def __init__(self): ...\n")

    if md.field:
        # We're using a dictionary here as a cheap order-preserving set, since Python dictionaries
        # preserve insertion order (which is important to us here).
        field_names = {}  # type: Dict[str, None]
        for fld in md.field:
            if fld.HasField("oneof_index"):
                field_names[md.oneof_decl[fld.oneof_index].name] = None
            field_names[fld.name] = None
        field_names_str = ", ".join(json.dumps(fld) for fld in field_names)
        args = f"self, field_name: _L[{field_names_str}]"
        buf.write(f"    def HasField({args}) -> _builtins.bool: ...\n")
        buf.write(f"    def ClearField({args}) -> None: ...\n")
    else:
        args = f"self, field_name: _typing.NoReturn"
        buf.write(f"    def HasField({args}) -> _typing.NoReturn: ...\n")
        buf.write(f"    def ClearField({args}) -> _typing.NoReturn: ...\n")

    if oneof_groups:
        for oneof_group_name, oneof_group_fields in oneof_groups.items():
            if len(oneof_groups) > 1:
                buf.write("    @_typing.overload\n")
            args = f'self, oneof_group: _L["{oneof_group_name}"]'
            ret = f"_L[{', '.join(json.dumps(f) if f else 'None' for f in oneof_group_fields)}]"
            buf.write(f"    def WhichOneof({args}) -> {ret}: ...\n")
    else:
        args = "self, oneof_group: _typing.NoReturn"
        ret = "_typing.NoReturn"
        buf.write(f"    def WhichOneof({args}) -> {ret}: ...\n")



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from asyncio import sleep
from datetime import datetime, timedelta
from typing import Awaitable, Callable, TypeVar

from grpc import RpcError, StatusCode

T = TypeVar("T")

DEFAULT_TIMEOUT = timedelta(minutes=1)


def is_retryable_exception(ex: Exception) -> bool:
    if not isinstance(ex, RpcError):
        return False

    # TODO: Expand on these categories as documented here:
    #  https://docs.daml.com/app-dev/grpc/error-codes.html
    return ex.code() == StatusCode.FAILED_PRECONDITION and ex.details().startswith(
        "PARTY_ALLOCATION_WITHOUT_CONNECTED_DOMAIN"
    )


async def retry(
    fn: Callable[[], Awaitable[T]],
    *,
    retry_safe_ex: Callable[[Exception], bool] = is_retryable_exception,
    timeout: timedelta = DEFAULT_TIMEOUT
) -> T:
    a, b = 0, 1
    start = datetime.utcnow()
    while True:
        try:
            return await fn()
        except Exception as ex:
            if (datetime.now() - start) <= timeout and retry_safe_ex(ex):
                # retry in increasing intervals according to the Fibonacci sequence;
                # exponential backoff generally makes us wait too long
                a, b = b, a + b
                await sleep(a)
            else:
                raise



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.duration_pb2 import Duration
from google.protobuf.internal.containers import RepeatedCompositeFieldContainer, RepeatedScalarFieldContainer
from google.protobuf.message import Message as _Message
from google.protobuf.timestamp_pb2 import Timestamp

from .value_pb2 import Identifier, Record, Value

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "Commands",
    "Command",
    "CreateCommand",
    "ExerciseCommand",
    "ExerciseByKeyCommand",
    "CreateAndExerciseCommand",
]


class Commands(_Message):
    ledger_id: _builtins.str
    workflow_id: _builtins.str
    application_id: _builtins.str
    command_id: _builtins.str
    party: _builtins.str
    @property
    def commands(self) -> RepeatedCompositeFieldContainer[Command]: ...
    @property
    def deduplication_time(self) -> Duration: ...
    @property
    def deduplication_duration(self) -> Duration: ...
    deduplication_offset: _builtins.str
    @property
    def min_ledger_time_abs(self) -> Timestamp: ...
    @property
    def min_ledger_time_rel(self) -> Duration: ...
    @property
    def act_as(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def read_as(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    submission_id: _builtins.str
    @_typing.overload
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., workflow_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., command_id: _typing.Optional[_builtins.str] = ..., party: _typing.Optional[_builtins.str] = ..., commands: _typing.Optional[_typing.Iterable[Command]] = ..., min_ledger_time_abs: _typing.Optional[Timestamp] = ..., min_ledger_time_rel: _typing.Optional[Duration] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., read_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ...): ...
    @_typing.overload
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., workflow_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., command_id: _typing.Optional[_builtins.str] = ..., party: _typing.Optional[_builtins.str] = ..., commands: _typing.Optional[_typing.Iterable[Command]] = ..., deduplication_time: Duration = ..., min_ledger_time_abs: _typing.Optional[Timestamp] = ..., min_ledger_time_rel: _typing.Optional[Duration] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., read_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ...): ...
    @_typing.overload
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., workflow_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., command_id: _typing.Optional[_builtins.str] = ..., party: _typing.Optional[_builtins.str] = ..., commands: _typing.Optional[_typing.Iterable[Command]] = ..., deduplication_duration: Duration = ..., min_ledger_time_abs: _typing.Optional[Timestamp] = ..., min_ledger_time_rel: _typing.Optional[Duration] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., read_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ...): ...
    @_typing.overload
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., workflow_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., command_id: _typing.Optional[_builtins.str] = ..., party: _typing.Optional[_builtins.str] = ..., commands: _typing.Optional[_typing.Iterable[Command]] = ..., deduplication_offset: _builtins.str = ..., min_ledger_time_abs: _typing.Optional[Timestamp] = ..., min_ledger_time_rel: _typing.Optional[Duration] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., read_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["ledger_id", "workflow_id", "application_id", "command_id", "party", "commands", "deduplication_period", "deduplication_time", "deduplication_duration", "deduplication_offset", "min_ledger_time_abs", "min_ledger_time_rel", "act_as", "read_as", "submission_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id", "workflow_id", "application_id", "command_id", "party", "commands", "deduplication_period", "deduplication_time", "deduplication_duration", "deduplication_offset", "min_ledger_time_abs", "min_ledger_time_rel", "act_as", "read_as", "submission_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["deduplication_period"]) -> _L[None, "deduplication_time", "deduplication_duration", "deduplication_offset"]: ...

class Command(_Message):
    @property
    def create(self) -> CreateCommand: ...
    @property
    def exercise(self) -> ExerciseCommand: ...
    @property
    def exerciseByKey(self) -> ExerciseByKeyCommand: ...
    @property
    def createAndExercise(self) -> CreateAndExerciseCommand: ...
    @_typing.overload
    def __init__(self): ...
    @_typing.overload
    def __init__(self, *, create: CreateCommand = ...): ...
    @_typing.overload
    def __init__(self, *, exercise: ExerciseCommand = ...): ...
    @_typing.overload
    def __init__(self, *, exerciseByKey: ExerciseByKeyCommand = ...): ...
    @_typing.overload
    def __init__(self, *, createAndExercise: CreateAndExerciseCommand = ...): ...
    def HasField(self, field_name: _L["command", "create", "exercise", "exerciseByKey", "createAndExercise"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["command", "create", "exercise", "exerciseByKey", "createAndExercise"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["command"]) -> _L[None, "create", "exercise", "exerciseByKey", "createAndExercise"]: ...

class CreateCommand(_Message):
    @property
    def template_id(self) -> Identifier: ...
    @property
    def create_arguments(self) -> Record: ...
    def __init__(self, *, template_id: _typing.Optional[Identifier] = ..., create_arguments: _typing.Optional[Record] = ...): ...
    def HasField(self, field_name: _L["template_id", "create_arguments"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["template_id", "create_arguments"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExerciseCommand(_Message):
    @property
    def template_id(self) -> Identifier: ...
    contract_id: _builtins.str
    choice: _builtins.str
    @property
    def choice_argument(self) -> Value: ...
    def __init__(self, *, template_id: _typing.Optional[Identifier] = ..., contract_id: _typing.Optional[_builtins.str] = ..., choice: _typing.Optional[_builtins.str] = ..., choice_argument: _typing.Optional[Value] = ...): ...
    def HasField(self, field_name: _L["template_id", "contract_id", "choice", "choice_argument"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["template_id", "contract_id", "choice", "choice_argument"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExerciseByKeyCommand(_Message):
    @property
    def template_id(self) -> Identifier: ...
    @property
    def contract_key(self) -> Value: ...
    choice: _builtins.str
    @property
    def choice_argument(self) -> Value: ...
    def __init__(self, *, template_id: _typing.Optional[Identifier] = ..., contract_key: _typing.Optional[Value] = ..., choice: _typing.Optional[_builtins.str] = ..., choice_argument: _typing.Optional[Value] = ...): ...
    def HasField(self, field_name: _L["template_id", "contract_key", "choice", "choice_argument"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["template_id", "contract_key", "choice", "choice_argument"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CreateAndExerciseCommand(_Message):
    @property
    def template_id(self) -> Identifier: ...
    @property
    def create_arguments(self) -> Record: ...
    choice: _builtins.str
    @property
    def choice_argument(self) -> Value: ...
    def __init__(self, *, template_id: _typing.Optional[Identifier] = ..., create_arguments: _typing.Optional[Record] = ..., choice: _typing.Optional[_builtins.str] = ..., choice_argument: _typing.Optional[Value] = ...): ...
    def HasField(self, field_name: _L["template_id", "create_arguments", "choice", "choice_argument"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["template_id", "create_arguments", "choice", "choice_argument"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
:mod:`dazl.model` package
=========================

This module is deprecated. These types have generally moved to :mod:`dazl.client` (for the API
introduced in dazl v5) or :mod:`dazl.protocols` (for the API introduced in dazl v8).

.. automodule:: dazl.model.core
.. automodule:: dazl.model.ledger
.. automodule:: dazl.model.lookup
.. automodule:: dazl.model.network
.. automodule:: dazl.model.reading
.. automodule:: dazl.model.types
.. automodule:: dazl.model.types_store
.. automodule:: dazl.model.writing

"""

import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from . import core, ledger, lookup, network, reading, types, writing

__all__ = ["core", "ledger", "lookup", "network", "reading", "writing"]

warnings.warn(
    "dazl.model is deprecated; these types have moved to either dazl.ledger or dazl.client.",
    DeprecationWarning,
    stacklevel=2,
)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains helper methods for detecting terminal capabilities.
"""

from subprocess import DEVNULL, PIPE, Popen
from typing import Optional, Tuple


def termsize() -> "Tuple[Optional[int], Optional[int]]":
    """
    Return the current size of the terminal. If the current terminal is not a tty, then
    ``(None, None)`` is returned.
    """
    try:
        with Popen(["stty", "size"], stdout=PIPE, stderr=DEVNULL) as proc:
            term_size_str = proc.stdout.read().decode("utf8") if proc.stdout is not None else None
    except FileNotFoundError:
        term_size_str = None
    if term_size_str is not None:
        dimensions = term_size_str.split()
        if len(dimensions) >= 2:
            return int(dimensions[0]), int(dimensions[1])
        else:
            return None, None
    else:
        return None, None


def print_termcap() -> None:
    """
    Output discovered settings to the terminal.
    """
    rows, cols = termsize()
    if rows is not None and cols is not None:
        print("The terminal size is:")
        print(termsize())
    else:
        print("Not a terminal.")


if __name__ == "__main__":
    print_termcap()



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ._base import RunLevel, validate_install_signal_handlers
from ._invoker import Invoker



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse

from .. import protoc, util
from ..syntax.python import rewrite_file_content
from ._header import HEADER

__all__ = ["run_plugin"]


def run_plugin(request: "CodeGeneratorRequest") -> "CodeGeneratorResponse":
    response = protoc.run_plugin("grpc_python", util.services_only(request))
    for file in response.file:
        file.content = rewrite_file_content(file.name, file.content)

    return util.with_file_header(response, HEADER)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import AbstractSet, Dict, Iterable, Iterator, List, Optional

from ...damlast.protocols import SymbolLookup
from ...prim import ContractData, ContractId, Party

__all__ = ["Formatter", "RowBuilder", "TableBuilder"]


class Formatter:
    """
    A formatter is a particular way of representing a list of rows that represent contracts visible
    by a set of parties.
    """

    def render(
        self, lookup: "SymbolLookup", parties: "AbstractSet[Party]", entries: "Iterable[RowBuilder]"
    ) -> "Iterator[str]":
        """
        Render the set of entries.

        :param lookup: The store that contains type information.
        :param parties: Set of parties to render.
        :param entries: A list of entries to render.
        :return: An asynchronous generator over string lines that make up the rendering.
        """
        raise NotImplementedError("render needs an implementation")


class TableBuilder:
    """
    State built up during a ledger run.
    """

    def __init__(self):
        self.entries = dict()  # type: Dict[ContractId, RowBuilder]

    def add(
        self,
        party: "Party",
        cid: "ContractId",
        cdata: "ContractData",
        time: "Optional[datetime]" = None,
    ) -> None:
        entry = self.entries.get(cid)
        if entry is not None:
            entry.extend(party, cdata)
        else:
            self.entries[cid] = RowBuilder(party, cid, cdata, time)

    def excluding_inactive(self) -> "Iterator[RowBuilder]":
        return filter(lambda e: not e.is_archived(), self)

    def filtered_by(self, predicate) -> "Iterator[RowBuilder]":
        return filter(predicate, self)

    def __iter__(self) -> "Iterator[RowBuilder]":
        return iter(self.entries.values())

    def __repr__(self) -> str:
        return "<LedgerCapture(entry_count={!r})>".format(len(self.entries))


class RowBuilder:
    """
    Represents a single "row" in the final output. This is a consolidation over the view that
    various parties see for a particular :class:`ContractId`.
    """

    def __init__(
        self, party: "Party", cid: "ContractId", cdata: "ContractData", time: "Optional[datetime]"
    ):
        self.parties = {party: True}  # type: Dict[Party, bool]
        self.cid = cid
        self.cdata = cdata
        self.time = time
        self.errors = []  # type: List[Exception]

    def extend(self, party: "Party", cdata: "Optional[ContractData]") -> None:
        self.parties[party] = cdata is not None

    def is_archived(self) -> bool:
        return all(not active for active in self.parties.values())

    def contract_state(self, party: "Party") -> "Optional[str]":
        p = self.parties.get(party)
        if p is not None:
            return "CREATED" if p else "ARCHIVED"
        else:
            return None

    def __repr__(self) -> str:
        return f"LedgerEntry(parties={self.parties}, cid={self.cid}, cdata={self.cdata})"



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from io import StringIO
from os.path import basename, dirname, splitext
from typing import DefaultDict, Set

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse

from ..syntax.python import all_decl
from ._header import HEADER

__all__ = ["run_plugin"]


def run_plugin(request: "CodeGeneratorRequest") -> "CodeGeneratorResponse":
    """
    Create `__init__.py` files for all directories.
    """
    directories = set()

    # for all of the requested files, add those files' parent directories, and the parents of all
    # THOSE directories
    for proto_dir in {dirname(f) for f in request.file_to_generate}:
        components = proto_dir.split("/")
        for i in range(0, len(components) + 1):
            directories.add("/".join(components[:i]))

    return CodeGeneratorResponse(
        file=[package_file(proto_dir, request) for proto_dir in sorted(directories)]
    )


def package_file(proto_dir: str, request: "CodeGeneratorRequest") -> "CodeGeneratorResponse.File":
    proto_package = proto_dir.replace("/", ".")

    import_map = defaultdict(set)  # type: DefaultDict[str, Set[str]]
    for file in request.proto_file:
        if file.package == proto_package:
            current_module_base = splitext(basename(file.name))[0]
            current_module_pb = "." + current_module_base + "_pb2"
            current_module_grpc = "." + current_module_base + "_pb2_grpc"
            for e in file.enum_type:
                import_map[current_module_pb].add(e.name)
            for m in file.message_type:
                import_map[current_module_pb].add(m.name)
            for s in file.service:
                import_map[current_module_grpc].add(s.name + "Stub")

    with StringIO() as buf:
        buf.write(HEADER)
        buf.write("\n")
        all_symbols = set()
        for f, i in import_map.items():
            buf.write(f"from {f} import {', '.join(sorted(i))}\n")
            all_symbols.update(i)
        buf.write("\n")
        buf.write(all_decl(sorted(all_symbols)))

        return CodeGeneratorResponse.File(name=f"{proto_dir}/__init__.py", content=buf.getvalue())



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer, RepeatedScalarFieldContainer
from google.protobuf.message import Message as _Message

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "GetParticipantIdRequest",
    "GetParticipantIdResponse",
    "GetPartiesRequest",
    "GetPartiesResponse",
    "ListKnownPartiesRequest",
    "ListKnownPartiesResponse",
    "AllocatePartyRequest",
    "AllocatePartyResponse",
    "PartyDetails",
]


class GetParticipantIdRequest(_Message):
    def __init__(self): ...
    def HasField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def ClearField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetParticipantIdResponse(_Message):
    participant_id: _builtins.str
    def __init__(self, *, participant_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["participant_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["participant_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetPartiesRequest(_Message):
    @property
    def parties(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    def __init__(self, *, parties: _typing.Optional[_typing.Iterable[_builtins.str]] = ...): ...
    def HasField(self, field_name: _L["parties"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["parties"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetPartiesResponse(_Message):
    @property
    def party_details(self) -> RepeatedCompositeFieldContainer[PartyDetails]: ...
    def __init__(self, *, party_details: _typing.Optional[_typing.Iterable[PartyDetails]] = ...): ...
    def HasField(self, field_name: _L["party_details"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["party_details"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListKnownPartiesRequest(_Message):
    def __init__(self): ...
    def HasField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def ClearField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListKnownPartiesResponse(_Message):
    @property
    def party_details(self) -> RepeatedCompositeFieldContainer[PartyDetails]: ...
    def __init__(self, *, party_details: _typing.Optional[_typing.Iterable[PartyDetails]] = ...): ...
    def HasField(self, field_name: _L["party_details"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["party_details"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class AllocatePartyRequest(_Message):
    party_id_hint: _builtins.str
    display_name: _builtins.str
    def __init__(self, *, party_id_hint: _typing.Optional[_builtins.str] = ..., display_name: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["party_id_hint", "display_name"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["party_id_hint", "display_name"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class AllocatePartyResponse(_Message):
    @property
    def party_details(self) -> PartyDetails: ...
    def __init__(self, *, party_details: _typing.Optional[PartyDetails] = ...): ...
    def HasField(self, field_name: _L["party_details"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["party_details"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class PartyDetails(_Message):
    party: _builtins.str
    display_name: _builtins.str
    is_local: _builtins.bool
    def __init__(self, *, party: _typing.Optional[_builtins.str] = ..., display_name: _typing.Optional[_builtins.str] = ..., is_local: _typing.Optional[_builtins.bool] = ...): ...
    def HasField(self, field_name: _L["party", "display_name", "is_local"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["party", "display_name", "is_local"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, AsyncIterable, Callable, Iterable, List, Tuple, TypeVar, Union, cast
from urllib.parse import urlparse

from grpc import (
    AuthMetadataContext,
    AuthMetadataPlugin,
    AuthMetadataPluginCallback,
    ChannelCredentials,
    composite_channel_credentials,
    metadata_call_credentials,
    ssl_channel_credentials,
)
from grpc.aio import (
    Channel,
    ClientCallDetails,
    StreamStreamCall,
    StreamStreamClientInterceptor,
    StreamUnaryCall,
    StreamUnaryClientInterceptor,
    UnaryStreamCall,
    UnaryStreamClientInterceptor,
    UnaryUnaryCall,
    UnaryUnaryClientInterceptor,
    insecure_channel,
    secure_channel,
)

from ..config import Config

__all__ = ["create_channel"]

RequestType = TypeVar("RequestType")
RequestIterableType = Union[Iterable[Any], AsyncIterable[Any]]
ResponseIterableType = AsyncIterable[Any]


def create_channel(config: "Config") -> "Channel":
    """
    Create a :class:`Channel` for the specified configuration.
    """
    u = urlparse(config.url.url)

    options = [
        ("grpc.max_send_message_length", -1),
        ("grpc.max_receive_message_length", -1),
    ]
    if not config.url.use_http_proxy:
        options.append(("grpc.enable_http_proxy", 0))

    if (u.scheme in ("https", "grpcs")) or config.ssl:
        if config.ssl.ca is None and config.ssl.cert is not None:
            credentials = ssl_channel_credentials(root_certificates=config.ssl.cert)
        else:
            credentials = ssl_channel_credentials(
                root_certificates=config.ssl.ca,
                private_key=config.ssl.cert_key,
                certificate_chain=config.ssl.cert,
            )
        if config.access.token_version is not None:
            # The grpc Credential objects do not actually define a formal interface, and are
            # used interchangeably in the code.
            #
            # Additionally there are some incorrect rules in the grpc-stubs typing rules that force
            # us to work around the type system.
            credentials = cast(
                ChannelCredentials,
                composite_channel_credentials(
                    credentials, metadata_call_credentials(GrpcAuth(config), name="auth gateway")
                ),
            )
        return secure_channel(u.netloc, credentials, tuple(options))

    elif config.access.token_version is not None:
        # Python/C++ libraries refuse to allow "credentials" objects to be passed around on
        # non-TLS channels, but they don't check interceptors; use an interceptor to inject
        # an Authorization header instead
        return insecure_channel(u.netloc, options, interceptors=[GrpcAuthInterceptor(config)])

    else:
        # no TLS, no tokens--simply create an insecure channel with no adornments
        return insecure_channel(u.netloc, options)


class GrpcAuth(AuthMetadataPlugin):
    def __init__(self, config: "Config"):
        self._config = config

    def __call__(self, context: "AuthMetadataContext", callback: "AuthMetadataPluginCallback"):
        # This overly verbose type signature is here to satisfy mypy and grpc-stubs
        options = []  # type: List[Tuple[str, Union[str, bytes]]]

        # TODO: Add support here for refresh tokens
        token = self._config.access.token
        if token:
            # note: gRPC headers MUST be lowercased
            options.append(("authorization", "Bearer " + self._config.access.token))

        callback(tuple(options), None)


class GrpcAuthInterceptor(
    UnaryUnaryClientInterceptor,
    UnaryStreamClientInterceptor,
    StreamUnaryClientInterceptor,
    StreamStreamClientInterceptor,
):
    """
    An interceptor that injects "Authorization" metadata into a request.

    This works around the fact that the C++ gRPC libraries (which Python is built on) highly
    discourage sending authorization data over the wire unless the connection is protected with TLS.
    """

    # NOTE: There are a number of typing errors in the grpc.aio classes, so we're ignoring a handful
    #  of lines until those problems are addressed.

    def __init__(self, config: "Config"):
        self._config = config

    async def intercept_unary_unary(
        self,
        continuation: "Callable[[ClientCallDetails, RequestType], UnaryUnaryCall]",
        client_call_details: ClientCallDetails,
        request: RequestType,
    ) -> "Union[UnaryUnaryCall, RequestType]":
        return await continuation(self._modify_client_call_details(client_call_details), request)

    async def intercept_unary_stream(
        self,
        continuation: "Callable[[ClientCallDetails, RequestType], UnaryStreamCall]",
        client_call_details: ClientCallDetails,
        request: RequestType,
    ) -> "Union[ResponseIterableType, UnaryStreamCall]":
        return await continuation(self._modify_client_call_details(client_call_details), request)

    async def intercept_stream_unary(
        self,
        continuation: "Callable[[ClientCallDetails, RequestType], StreamUnaryCall]",
        client_call_details: ClientCallDetails,
        request_iterator: RequestIterableType,
    ) -> StreamUnaryCall:
        return await continuation(
            self._modify_client_call_details(client_call_details), request_iterator  # type: ignore
        )

    async def intercept_stream_stream(
        self,
        continuation: Callable[[ClientCallDetails, RequestType], StreamStreamCall],
        client_call_details: ClientCallDetails,
        request_iterator: RequestIterableType,
    ) -> "Union[ResponseIterableType, StreamStreamCall]":
        return await continuation(
            self._modify_client_call_details(client_call_details), request_iterator  # type: ignore
        )

    def _modify_client_call_details(self, client_call_details: ClientCallDetails):
        if (
            "authorization" not in client_call_details.metadata
            and self._config.access.token_version is not None
        ):
            client_call_details.metadata.add("authorization", f"Bearer {self._config.access.token}")

        return client_call_details



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/metering_report_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n:com/daml/ledger/api/v1/admin/metering_report_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin\x1a\x1fgoogle/protobuf/timestamp.proto\"\x9d\x01\n\x18GetMeteringReportRequest\x12.\n\x04\x66rom\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x04\x66rom\x12*\n\x02to\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x02to\x12%\n\x0e\x61pplication_id\x18\x03 \x01(\tR\rapplicationId\"\xa7\x02\n\x19GetMeteringReportResponse\x12P\n\x07request\x18\x01 \x01(\x0b\x32\x36.com.daml.ledger.api.v1.admin.GetMeteringReportRequestR\x07request\x12\x66\n\x12participant_report\x18\x02 \x01(\x0b\x32\x37.com.daml.ledger.api.v1.admin.ParticipantMeteringReportR\x11participantReport\x12P\n\x16report_generation_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x14reportGenerationTime\"\xc7\x01\n\x19ParticipantMeteringReport\x12%\n\x0eparticipant_id\x18\x01 \x01(\tR\rparticipantId\x12\x19\n\x08is_final\x18\x02 \x01(\x08R\x07isFinal\x12h\n\x13\x61pplication_reports\x18\x03 \x03(\x0b\x32\x37.com.daml.ledger.api.v1.admin.ApplicationMeteringReportR\x12\x61pplicationReports\"c\n\x19\x41pplicationMeteringReport\x12%\n\x0e\x61pplication_id\x18\x01 \x01(\tR\rapplicationId\x12\x1f\n\x0b\x65vent_count\x18\x02 \x01(\x03R\neventCount2\x9e\x01\n\x15MeteringReportService\x12\x84\x01\n\x11GetMeteringReport\x12\x36.com.daml.ledger.api.v1.admin.GetMeteringReportRequest\x1a\x37.com.daml.ledger.api.v1.admin.GetMeteringReportResponseB\xa4\x01\n\x1c\x63om.daml.ledger.api.v1.adminB\x18MeteringReportOuterClassZKgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/admin\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3')



_GETMETERINGREPORTREQUEST = DESCRIPTOR.message_types_by_name['GetMeteringReportRequest']
_GETMETERINGREPORTRESPONSE = DESCRIPTOR.message_types_by_name['GetMeteringReportResponse']
_PARTICIPANTMETERINGREPORT = DESCRIPTOR.message_types_by_name['ParticipantMeteringReport']
_APPLICATIONMETERINGREPORT = DESCRIPTOR.message_types_by_name['ApplicationMeteringReport']
GetMeteringReportRequest = _reflection.GeneratedProtocolMessageType('GetMeteringReportRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETERINGREPORTREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.metering_report_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetMeteringReportRequest)
  })
_sym_db.RegisterMessage(GetMeteringReportRequest)

GetMeteringReportResponse = _reflection.GeneratedProtocolMessageType('GetMeteringReportResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMETERINGREPORTRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.metering_report_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetMeteringReportResponse)
  })
_sym_db.RegisterMessage(GetMeteringReportResponse)

ParticipantMeteringReport = _reflection.GeneratedProtocolMessageType('ParticipantMeteringReport', (_message.Message,), {
  'DESCRIPTOR' : _PARTICIPANTMETERINGREPORT,
  '__module__' : 'com.daml.ledger.api.v1.admin.metering_report_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ParticipantMeteringReport)
  })
_sym_db.RegisterMessage(ParticipantMeteringReport)

ApplicationMeteringReport = _reflection.GeneratedProtocolMessageType('ApplicationMeteringReport', (_message.Message,), {
  'DESCRIPTOR' : _APPLICATIONMETERINGREPORT,
  '__module__' : 'com.daml.ledger.api.v1.admin.metering_report_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ApplicationMeteringReport)
  })
_sym_db.RegisterMessage(ApplicationMeteringReport)

_METERINGREPORTSERVICE = DESCRIPTOR.services_by_name['MeteringReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.daml.ledger.api.v1.adminB\030MeteringReportOuterClassZKgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/admin\252\002\034Com.Daml.Ledger.Api.V1.Admin'
  _GETMETERINGREPORTREQUEST._serialized_start=126
  _GETMETERINGREPORTREQUEST._serialized_end=283
  _GETMETERINGREPORTRESPONSE._serialized_start=286
  _GETMETERINGREPORTRESPONSE._serialized_end=581
  _PARTICIPANTMETERINGREPORT._serialized_start=584
  _PARTICIPANTMETERINGREPORT._serialized_end=783
  _APPLICATIONMETERINGREPORT._serialized_start=785
  _APPLICATIONMETERINGREPORT._serialized_end=884
  _METERINGREPORTSERVICE._serialized_start=887
  _METERINGREPORTSERVICE._serialized_end=1045
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse

__all__ = ["with_file_header", "services_only"]


def with_file_header(response: "CodeGeneratorResponse", header: str) -> "CodeGeneratorResponse":
    """
    Add a file header to each file in the response, and return a new :class:`CodeGeneratorResponse`
    with the modified files.

    :param response: The :class:`CodeGeneratorResponse` that contains files to modify.
    :param header: The header string to prepend to each file.
    :return: A new :class:`CodeGeneratorResponse`.
    """
    return CodeGeneratorResponse(
        file=[
            CodeGeneratorResponse.File(name=f.name, content=header + f.content)
            for f in response.file
        ]
    )


def services_only(request: "CodeGeneratorRequest") -> "CodeGeneratorRequest":
    """
    Create a :class:`CodeGeneratorRequest` with the non-gRPC files stripped out (those without
    ``service`` declarations. The original request is not modified.
    """
    # include all of the Protobuf files, but restrict the list of "files_to_generate"
    files_with_services = {f.name for f in request.proto_file if len(f.service) > 0}
    return CodeGeneratorRequest(
        proto_file=request.proto_file,
        file_to_generate=[f for f in request.file_to_generate if f in files_with_services],
    )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ._basic import all_decl
from .imports import ImportContext, rewrite_file_content
from .pb import py_message_package, py_scalar_type, py_service_package
from .symbols import SymbolTable, Usage
from .types import PyType

__all__ = [
    "all_decl",
    "ImportContext",
    "rewrite_file_content",
    "py_message_package",
    "py_service_package",
    "py_scalar_type",
    "SymbolTable",
    "PyType",
    "Usage",
]



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, typing as _typing

import grpc as _grpc
from grpc import aio as _grpc_aio

from .package_management_service_pb2 import ListKnownPackagesRequest, ListKnownPackagesResponse, UploadDarFileRequest, UploadDarFileResponse

__all__ = [
    "PackageManagementServiceStub",
]


# noinspection PyPep8Naming,DuplicatedCode
class PackageManagementServiceStub:
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc.Channel) -> _PackageManagementServiceBlockingStub: ...  # type: ignore
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc_aio.Channel) -> _PackageManagementServiceAsyncStub: ...  # type: ignore
    def ListKnownPackages(self, __1: ListKnownPackagesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[ListKnownPackagesResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, ListKnownPackagesResponse]]: ...
    def UploadDarFile(self, __1: UploadDarFileRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[UploadDarFileResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, UploadDarFileResponse]]: ...

# noinspection PyPep8Naming,DuplicatedCode
class _PackageManagementServiceBlockingStub(PackageManagementServiceStub):
    def ListKnownPackages(self, __1: ListKnownPackagesRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> ListKnownPackagesResponse: ...
    def UploadDarFile(self, __1: UploadDarFileRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> UploadDarFileResponse: ...

# noinspection PyPep8Naming,DuplicatedCode
class _PackageManagementServiceAsyncStub(PackageManagementServiceStub):
    def ListKnownPackages(self, __1: ListKnownPackagesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, ListKnownPackagesResponse]: ...
    def UploadDarFile(self, __1: UploadDarFileRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, UploadDarFileResponse]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping as MappingABC
from typing import Any, Callable, Collection, Dict, Mapping, Optional, Union

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.lookup import parse_type_con_name
from ..prim import ContractData

__all__ = ["ContractMatch", "is_match", "Query", "Queries", "parse_query", "Filter", "EMPTY"]

ContractMatch = Union[None, Callable[[ContractData], bool], ContractData]

TemplateName = Union[str, "TypeConName"]
Query = Union[None, ContractData, Callable[[ContractData], bool]]
Queries = Union[None, TemplateName, Collection[TemplateName], Mapping[TemplateName, Query]]


class Filter:
    server_side: Optional[ContractData]
    client_side: Optional[Callable[[ContractData], bool]]

    def __init__(
        self,
        *,
        server_side: Optional[ContractData] = None,
        client_side: Optional[Callable[[ContractData], bool]] = None
    ):
        self.server_side = server_side
        self.client_side = client_side

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Filter):
            return False

        return self.server_side == other.server_side and self.client_side == other.client_side


EMPTY = Filter()


def parse_query(*q: Queries, server_side_filters: bool) -> "Optional[Mapping[TypeConName, Filter]]":
    """
    Take a :class:`Queries` object and convert it to a mapping of template names to :class:`Filter`.

    :param q:
        Either:
        * omitted (this means a match on all templates)
        * a single TemplateName (str or TypeConName)
        * a collection of TemplateNames
        * a mapping of TemplateNames to:
            * query objects as defined in https://docs.daml.com/json-api/search-query-language.html
            * functions that perform client-side filtering of contract data
    :param server_side_filters:
        ``True`` to possibly create :class:`Filter` objects with a ``Filter.server_side`` property
        defined (as defined in https://docs.daml.com/json-api/search-query-language.html); ``False``
        to implement all filters as client-side filters.
    :return:
        ``None`` if _all_ templates are to be matched; otherwise a mapping of :class:`TypeConName`
        to :class:`Filter` objects.
    """
    tq = None  # type: Optional[Dict[TypeConName, Filter]]
    for query in q:
        if query is not None and query != "*" and query != "*:*":
            if tq is None:
                tq = {}

            tn = {}  # type: Dict[TypeConName, Filter]
            if isinstance(query, str):
                tn = {parse_type_con_name(query): EMPTY}
            elif isinstance(query, TypeConName):
                tn = {query: EMPTY}
            elif isinstance(query, MappingABC):
                for key, value in query.items():
                    if isinstance(key, str):
                        key = parse_type_con_name(key)
                    tn[key] = _parse_query_filter(value, server_side_filters=server_side_filters)
            else:
                for key in query:
                    if isinstance(key, str):
                        key = parse_type_con_name(key)
                    tn[key] = EMPTY

            tq.update(tn)

    return tq


def _parse_query_filter(__tq: Query, *, server_side_filters: bool) -> Filter:
    if not __tq:
        return EMPTY
    if callable(__tq):
        return Filter(client_side=__tq)
    if server_side_filters:
        return Filter(server_side=__tq)
    else:
        return Filter(client_side=lambda cdata: is_match(__tq, cdata))


def is_match(predicate: ContractMatch, cdata: ContractData) -> bool:
    """
    Determine whether a contract matches a predicate expression.
    """
    if predicate is None:
        return True

    if callable(predicate):
        return predicate(cdata)

    for key, predicate_value in predicate.items():
        cdata_value = cdata.get(key)
        if cdata_value is None:
            # None only matches with None
            if predicate_value is not None:
                return False

        elif callable(predicate_value):
            # support lambdas as predicate values
            if not predicate_value(cdata_value):
                return False

        elif isinstance(predicate_value, dict):
            # deep dictionaries are for variants or records
            return is_match(predicate_value, cdata_value)

        elif predicate_value != cdata_value:
            # exact matches are, of course, supported
            return False

        else:
            # TODO: think of convenient ways to deal with lists
            pass

    return True



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Collection, NewType, Sequence, Union, overload

from .basic import to_str

__all__ = ["Party", "to_party", "to_parties"]

Party = NewType("Party", str)


def to_party(o: "Any") -> "Party":
    return Party(to_str(o))


@overload
def to_parties(__object: None) -> None:
    ...


@overload
def to_parties(
    __object: "Union[str, Party, Collection[str], Collection[Party]]",
) -> "Sequence[Party]":
    ...


def to_parties(
    __object: "Union[None, str, Party, Collection[str], Collection[Party]]",
) -> "Union[None, Sequence[Party]]":
    """
    Return the specified object as a collection of parties.

    :param __object:
        The object to convert to a set of parties.
    """
    if __object is None:
        return None
    elif isinstance(__object, str):
        return (Party(__object),)
    else:
        return tuple(Party(p) for p in __object)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the Python API for interacting with the Ledger API.
"""

__all__ = [
    "AIOPartyClient",
    "Command",
    "connect",
    "ContractData",
    "ContractId",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "DazlError",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "LOG",
    "Network",
    "Party",
    "SimplePartyClient",
    "__version__",
    "async_network",
    "create",
    "create_and_exercise",
    "exercise",
    "exercise_by_key",
    "frozendict",
    "run",
    "setup_default_logger",
    "simple_client",
    "write_acs",
]

from ._logging import LOG
from .client import AIOPartyClient, Network, SimplePartyClient, async_network, run, simple_client
from .client.commands import (
    CreateAndExerciseCommand,
    CreateCommand,
    ExerciseByKeyCommand,
    ExerciseCommand,
    create,
    create_and_exercise,
    exercise,
    exercise_by_key,
)
from .ledger import Command, connect
from .model.core import ContractId
from .pretty.table import write_acs
from .prim import ContractData, DazlError, FrozenDict as frozendict, Party
from .util.logging import setup_default_logger

# Large Protobuf message support

# These methods are undocumented, but is required to read even moderately sized DALF's.
try:
    # For Protobuf libraries version 4 or later (using upb)
    # noinspection PyPackageRequirements,PyUnresolvedReferences,PyProtectedMember
    from google._upb._message import SetAllowOversizeProtos

    SetAllowOversizeProtos(True)
except ImportError:
    pass

try:
    # For Protobuf libraries version 3 or earlier (using upb)
    # noinspection PyPackageRequirements,PyUnresolvedReferences,PyProtectedMember
    from google.protobuf.pyext import _message

    _message.SetAllowOversizeProtos(True)
except ImportError:
    pass


__version__ = "7.9.0"



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
DAML-LF Archive parsing
-----------------------

Functions for creating :class:`Archive` and :class:`Package` instances.

:func:`parse_archive` can be used to take an arbitrary byte payload from the Package Service and
convert it to an :class:`Archive`; an :class:`Archive` contains the package and its modules, which
in turn contain templates, data types, and values.
"""

import sys
import time
from typing import Optional

from .._gen.com.daml.daml_lf_1_14.daml_lf_pb2 import ArchivePayload
from .daml_lf_1 import Archive, PackageRef
from .pb_parse import ProtobufParser

__all__ = ["parse_archive", "parse_archive_payload"]


def parse_archive(package_id: "PackageRef", archive_bytes: bytes) -> "Archive":
    """
    Convert ``bytes`` into an :class:`Archive`.
    """
    archive_pb = parse_archive_payload(package_id, archive_bytes)

    parser = ProtobufParser(package_id)
    package = parser.parse_Package(archive_pb.daml_lf_1)

    return Archive(package_id, package)


def parse_archive_payload(
    package_id: "Optional[PackageRef]", archive_bytes: bytes
) -> "ArchivePayload":
    """
    Convert ``bytes`` into a :class:`G.ArchivePayload`.

    Note that this function will temporarily increase Python's recursion limit to handle cases where
    parsing a DAML-LF archive requires deeper recursion limits.
    """
    # noinspection PyPackageRequirements
    from google.protobuf.message import DecodeError

    from .. import LOG

    current_time = time.time()

    prev_recursion_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(5000)
    archive_payload = ArchivePayload()
    try:
        archive_payload.ParseFromString(archive_bytes)
    except DecodeError:
        # noinspection PyPackageRequirements
        from google.protobuf.internal import api_implementation  # type: ignore

        if api_implementation.Type() == "cpp":
            LOG.error("Failed to decode metadata. This may be due to bugs in the native Protobuf")
            LOG.error("implementation as exposed through Python, so setting an environment")
            LOG.error("variable to force a non-native implementation may help work around this")
            LOG.error("problem:")
            LOG.error("    export PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION=python")
        raise
    finally:
        sys.setrecursionlimit(prev_recursion_limit)

    final_time = time.time()
    total_millis = (final_time - current_time) * 1000
    if package_id is None:
        LOG.info("Parsed %s bytes of metadata in %0.2f ms.", len(archive_bytes), total_millis)
    else:
        LOG.info(
            "Parsed %s bytes of metadata (package ID %r) in %0.2f ms.",
            len(archive_bytes),
            package_id,
            total_millis,
        )

    return archive_payload



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/version_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import experimental_features_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_experimental__features__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,com/daml/ledger/api/v1/version_service.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\x32\x63om/daml/ledger/api/v1/experimental_features.proto\"9\n\x1aGetLedgerApiVersionRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\"\x7f\n\x1bGetLedgerApiVersionResponse\x12\x18\n\x07version\x18\x01 \x01(\tR\x07version\x12\x46\n\x08\x66\x65\x61tures\x18\x02 \x01(\x0b\x32*.com.daml.ledger.api.v1.FeaturesDescriptorR\x08\x66\x65\x61tures\"\xbe\x01\n\x12\x46\x65\x61turesDescriptor\x12V\n\x0fuser_management\x18\x02 \x01(\x0b\x32-.com.daml.ledger.api.v1.UserManagementFeatureR\x0euserManagement\x12P\n\x0c\x65xperimental\x18\x01 \x01(\x0b\x32,.com.daml.ledger.api.v1.ExperimentalFeaturesR\x0c\x65xperimental\"\x93\x01\n\x15UserManagementFeature\x12\x1c\n\tsupported\x18\x01 \x01(\x08R\tsupported\x12-\n\x13max_rights_per_user\x18\x02 \x01(\x05R\x10maxRightsPerUser\x12-\n\x13max_users_page_size\x18\x03 \x01(\x05R\x10maxUsersPageSize2\x90\x01\n\x0eVersionService\x12~\n\x13GetLedgerApiVersion\x12\x32.com.daml.ledger.api.v1.GetLedgerApiVersionRequest\x1a\x33.com.daml.ledger.api.v1.GetLedgerApiVersionResponseB\x92\x01\n\x16\x63om.daml.ledger.api.v1B\x18VersionServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_GETLEDGERAPIVERSIONREQUEST = DESCRIPTOR.message_types_by_name['GetLedgerApiVersionRequest']
_GETLEDGERAPIVERSIONRESPONSE = DESCRIPTOR.message_types_by_name['GetLedgerApiVersionResponse']
_FEATURESDESCRIPTOR = DESCRIPTOR.message_types_by_name['FeaturesDescriptor']
_USERMANAGEMENTFEATURE = DESCRIPTOR.message_types_by_name['UserManagementFeature']
GetLedgerApiVersionRequest = _reflection.GeneratedProtocolMessageType('GetLedgerApiVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETLEDGERAPIVERSIONREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.version_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetLedgerApiVersionRequest)
  })
_sym_db.RegisterMessage(GetLedgerApiVersionRequest)

GetLedgerApiVersionResponse = _reflection.GeneratedProtocolMessageType('GetLedgerApiVersionResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETLEDGERAPIVERSIONRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.version_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetLedgerApiVersionResponse)
  })
_sym_db.RegisterMessage(GetLedgerApiVersionResponse)

FeaturesDescriptor = _reflection.GeneratedProtocolMessageType('FeaturesDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESDESCRIPTOR,
  '__module__' : 'com.daml.ledger.api.v1.version_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.FeaturesDescriptor)
  })
_sym_db.RegisterMessage(FeaturesDescriptor)

UserManagementFeature = _reflection.GeneratedProtocolMessageType('UserManagementFeature', (_message.Message,), {
  'DESCRIPTOR' : _USERMANAGEMENTFEATURE,
  '__module__' : 'com.daml.ledger.api.v1.version_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.UserManagementFeature)
  })
_sym_db.RegisterMessage(UserManagementFeature)

_VERSIONSERVICE = DESCRIPTOR.services_by_name['VersionService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\030VersionServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _GETLEDGERAPIVERSIONREQUEST._serialized_start=124
  _GETLEDGERAPIVERSIONREQUEST._serialized_end=181
  _GETLEDGERAPIVERSIONRESPONSE._serialized_start=183
  _GETLEDGERAPIVERSIONRESPONSE._serialized_end=310
  _FEATURESDESCRIPTOR._serialized_start=313
  _FEATURESDESCRIPTOR._serialized_end=503
  _USERMANAGEMENTFEATURE._serialized_start=506
  _USERMANAGEMENTFEATURE._serialized_end=653
  _VERSIONSERVICE._serialized_start=656
  _VERSIONSERVICE._serialized_end=800
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, typing as _typing

import grpc as _grpc
from grpc import aio as _grpc_aio

from .package_service_pb2 import GetPackageRequest, GetPackageResponse, GetPackageStatusRequest, GetPackageStatusResponse, ListPackagesRequest, ListPackagesResponse

__all__ = [
    "PackageServiceStub",
]


# noinspection PyPep8Naming,DuplicatedCode
class PackageServiceStub:
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc.Channel) -> _PackageServiceBlockingStub: ...  # type: ignore
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc_aio.Channel) -> _PackageServiceAsyncStub: ...  # type: ignore
    def ListPackages(self, __1: ListPackagesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[ListPackagesResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, ListPackagesResponse]]: ...
    def GetPackage(self, __1: GetPackageRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetPackageResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetPackageResponse]]: ...
    def GetPackageStatus(self, __1: GetPackageStatusRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetPackageStatusResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetPackageStatusResponse]]: ...

# noinspection PyPep8Naming,DuplicatedCode
class _PackageServiceBlockingStub(PackageServiceStub):
    def ListPackages(self, __1: ListPackagesRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> ListPackagesResponse: ...
    def GetPackage(self, __1: GetPackageRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetPackageResponse: ...
    def GetPackageStatus(self, __1: GetPackageStatusRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetPackageStatusResponse: ...

# noinspection PyPep8Naming,DuplicatedCode
class _PackageServiceAsyncStub(PackageServiceStub):
    def ListPackages(self, __1: ListPackagesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, ListPackagesResponse]: ...
    def GetPackage(self, __1: GetPackageRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetPackageResponse]: ...
    def GetPackageStatus(self, __1: GetPackageStatusRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetPackageStatusResponse]: ...
